# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "26/10/2021"


import copy
import functools
import logging
import os
from typing import Iterable, Optional

from nabu import version as nabu_version
from nabu.pipeline.config import generate_nabu_configfile
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from processview.core.manager.manager import ProcessManager, DatasetState
from processview.core.superviseprocess import SuperviseProcess

from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as open_hdf5
from tomwer.core.utils.deprecation import deprecated_warning

from tomwer.core.cluster.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.process.reconstruction.nabu.plane import NabuPlane
from tomwer.core.process.task import Task
from tomwer.core.process.icat.gallery import (
    IcatScreenshots,
    deduce_dataset_gallery_location,
    select_screenshot_from_volume,
)
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.dictutils import concatenate_dict
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.io.utils.h5pyutils import EntryReader
from tomwer.utils import docstring
from tomwer.io.utils import format_stderr_stdout

from . import settings, utils
from .nabucommon import ResultsLocalRun, ResultSlurmRun, _NabuBaseReconstructor
from .target import Target

_logger = logging.getLogger(__name__)


def run_volume_reconstruction(
    scan: TomwerScanBase,
    config: dict,
    dry_run: bool,
    process_id: Optional[int] = None,
) -> tuple:
    """
    Run a volume reconstruction. Scan need to have reconstruction parameters for nabu.

    Behavior: will clear link to the last volume reconstructed

    :param int process_id: optional process id

    :return: succeed, stdouts, stderrs, configs, future_scan
    :rtype: tuple
    """

    if scan.nabu_recons_params in ({}, None):
        raise ValueError(
            "no configuration provided. You should run a "
            "reconstruction from nabuslices first."
        )
    cluster_config = config.pop("cluster_config", None)
    if cluster_config == {}:
        cluster_config = None
    elif isinstance(cluster_config, SlurmClusterConfiguration):
        cluster_config = cluster_config.to_dict()

    if cluster_config is None:
        target = Target.LOCAL
    else:
        target = Target.SLURM

    # # beam shape is not directly used by nabu (uses ctf_geometry directly)
    # config.get("phase", {}).pop("beam_shape", None)

    # config_volume = copy.copy(config)
    # config_nabu_slices = copy.deepcopy(scan.nabu_recons_params)
    # if "tomwer_slices" in config_nabu_slices:
    #     del config_nabu_slices["tomwer_slices"]

    # if "phase" in config_nabu_slices and "delta_beta" in config_nabu_slices["phase"]:
    #     pag_dbs = config_nabu_slices["phase"]["delta_beta"]
    #     if isinstance(pag_dbs, str):
    #         pag_dbs = utils.retrieve_lst_of_value_from_str(pag_dbs, type_=float)
    #     if len(pag_dbs) > 1:
    #         raise ValueError(
    #             "Several value of delta / beta found for volume reconstruction"
    #         )
    # scan.clear_latest_vol_reconstructions()

    if process_id is not None:
        try:
            process_name = ProcessManager().get_process(process_id=process_id).name
        except KeyError:
            process_name = "unknow"
    else:
        process_name = ""

    volume_reconstructor = VolumeRunner(
        scan=scan,
        config_nabu=config,
        cluster_config=cluster_config,
        dry_run=dry_run,
        target=target,
        process_name=process_name,
    )
    try:
        results = volume_reconstructor.run()
    except TimeoutError as e:
        _logger.error(e)
        return None
    else:
        assert len(results) == 1, "only one volume should be reconstructed"
        res = results[0]
        # tag latest reconstructions
        if isinstance(res, ResultsLocalRun) and res.results_identifiers is not None:
            scan.set_latest_vol_reconstructions(res.results_identifiers)
        # create future if needed
        if isinstance(res, ResultSlurmRun):
            future_tomo_obj = FutureTomwerObject(
                tomo_obj=scan,
                futures=tuple(res.future_slurm_jobs),
                process_requester_id=process_id,
            )

        else:
            future_tomo_obj = None
        succeed = res.success
        stdouts = (
            [
                res.std_out,
            ]
            if hasattr(res, "std_out")
            else []
        )
        stderrs = (
            [
                res.std_err,
            ]
            if hasattr(res, "std_err")
            else []
        )
        configs = (
            [
                res.config,
            ]
            if res is not None
            else []
        )

        return succeed, stdouts, stderrs, configs, future_tomo_obj


class VolumeRunner(_NabuBaseReconstructor):
    """
    Class used to reconstruct a full volume with Nabu.
    Locally or on a cluster.
    """

    EXPECTS_SINGLE_SLICE = False

    def __init__(
        self,
        scan: TomwerScanBase,
        config_nabu,
        cluster_config: Optional[dict],
        dry_run: bool,
        target: Target,
        process_name: str,
    ) -> None:
        super().__init__(
            scan=scan,
            dry_run=dry_run,
            target=target,
            cluster_config=cluster_config,
            process_name=process_name,
        )
        self._config = config_nabu

    @property
    def configuration(self):
        return self._config

    @docstring(_NabuBaseReconstructor)
    def run(self) -> Iterable:
        dataset_params = self.scan.get_nabu_dataset_info()
        if "dataset" in self.configuration:
            dataset_params.update(self.configuration["dataset"])
        self.configuration["dataset"] = dataset_params
        self.configuration["resources"] = utils.get_nabu_resources_desc(
            scan=self.scan, workers=1, method="local"
        )

        # force overwrite results
        if "output" not in self.configuration:
            self.configuration["output"] = {}
        config_slices, cfg_folder = self._treateOutputConfig(self.configuration)

        # force overwriting results
        config_slices["output"].update({"overwrite_results": 1})

        # check and clamp `start_z` and `end_z`
        if "reconstruction" in self.configuration:
            for key in ("start_z", "end_z"):
                value = config_slices["reconstruction"].get(key)
                if value is None:
                    continue

                value = int(value)
                if self.scan.dim_2 is not None and value >= self.scan.dim_2:
                    _logger.warning(
                        f"{key} > max_size (radio height: {self.scan.dim_2}). Set it to -1 (maximum)"
                    )
                    value = -1
                config_slices["reconstruction"][key] = value

        name = (
            config_slices["output"]["file_prefix"] + settings.NABU_CONFIG_FILE_EXTENSION
        )
        if not isinstance(self.scan, EDFTomoScan):
            name = "_".join((self.scan.entry.lstrip("/"), name))
        conf_file = os.path.join(cfg_folder, name)
        _logger.info(f"{self.scan}: create {conf_file}")

        # make sure output location exists
        os.makedirs(config_slices["output"]["location"], exist_ok=True)

        # add some tomwer metadata and save the configuration
        # note: for now the section is ignored by nabu but shouldn't stay that way
        with utils.TomwerInfo(config_slices) as config_to_dump:
            generate_nabu_configfile(
                conf_file,
                nabu_fullfield_default_config,
                config=config_to_dump,
                options_level="advanced",
            )
            return tuple(
                [
                    self._process_config(
                        config_to_dump=config_to_dump,
                        config_file=conf_file,
                        info="nabu volume reconstruction",
                        file_format=config_slices["output"]["file_format"],
                        process_name=self.process_name,
                    )
                ]
            )

    @docstring(_NabuBaseReconstructor)
    def _get_futures_slurm_callback(self, config_to_dump) -> tuple:
        # add callback to set slices reconstructed urls
        class CallBack:
            # we cannot create a future directly because distributed enforce
            # the callback to have a function signature with only the future
            # as single parameter.
            def __init__(self, f_partial, scan) -> None:
                self.f_partial = f_partial
                self.scan = scan

            def process(self, fn):
                if fn.done() and not (fn.cancelled() or fn.exception()):
                    # update reconstruction urls only if processing succeed.
                    recons_identifiers = self.f_partial()
                    self.scan.add_latest_vol_reconstructions(recons_identifiers)

        file_format = config_to_dump["output"]["file_format"]
        callback = functools.partial(
            utils.get_recons_volume_identifier,
            file_prefix=config_to_dump["output"]["file_prefix"],
            location=config_to_dump["output"]["location"],
            file_format=file_format,
            scan=self.scan,
            slice_index=None,
            axis=NabuPlane.XY,  # for volume we always reconstruct along XY plane
        )

        return (CallBack(callback, self.scan),)

    def _treateOutputConfig(self, config) -> tuple:
        """

        :return: (nabu config dict, nabu extra options)
        """
        config = copy.deepcopy(config)
        config, nabu_cfg_folder = super()._treateOutputSliceConfig(config)
        os.makedirs(config["output"]["location"], exist_ok=True)

        # adapt config_s to specific volume treatment
        if "postproc" in config:
            config["postproc"] = config["postproc"]

        # make sure start_[x] and end_[x] come from config
        for key in ("start_x", "end_x", "start_y", "end_y", "start_z", "end_z"):
            if key in config:
                config["reconstruction"][key] = config[key]
                del config[key]

        return config, nabu_cfg_folder

    @docstring(_NabuBaseReconstructor)
    def _get_file_basename_reconstruction(self, pag, db, ctf, axis):
        """

        :param TomwerScanBase scan: scan reconstructed
        :param Union[None, int] slice_index: index of the slice reconstructed.
                                            if None, we want to reconstruct the
                                            entire volume
        :param bool pag: is it a paganin reconstruction
        :param int db: delta / beta parameter
        :param axis: axis over which the reconstruction goes. For volume always expected to be z. So ignored in the function
        :return: basename of the file reconstructed (without any extension)
        """
        assert type(db) in (int, type(None))
        assert not pag == ctf == True, "cannot ask for both pag and ctf active"
        if isinstance(self.scan, NXtomoScan):
            basename, _ = os.path.splitext(self.scan.master_file)
            basename = os.path.basename(basename)
            try:
                # if there is more than one entry in the file append the entry name to the file basename
                with open_hdf5(self.scan.master_file) as h5f:
                    if len(h5f.keys()) > 1:
                        basename = "_".join((basename, self.scan.entry.strip("/")))
            except Exception:
                pass
        else:
            basename = os.path.basename(self.scan.path)

        if pag:
            return "_".join((basename + "pag", "db" + str(db).zfill(4), "vol"))
        elif ctf:
            return "_".join((basename + "ctf", "db" + str(db).zfill(4), "vol"))
        else:
            return "_".join((basename, "vol"))


class NabuVolumeTask(
    Task,
    SuperviseProcess,
    input_names=("data", "nabu_params"),
    output_names=("data", "volumes", "future_tomo_obj", "screenshots"),
    optional_input_names=(
        "dry_run",
        "nabu_extra_params",  # some parameter that must update 'nabu_params' before launching the reconstruction. Such as z range...
        "serialize_output_data",
    ),
):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        SuperviseProcess.__init__(self, process_id=process_id)
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        self._dry_run = inputs.get("dry_run", False)
        self._current_processing = None

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            self.outputs.data = None
            return

        # update scan reconstruction parameters used
        nabu_params = concatenate_dict(
            copy.deepcopy(self.inputs.nabu_params),
            self.get_input_value("nabu_extra_params", dict()),
        )
        scan.nabu_recons_params = nabu_params
        scan.clear_latest_vol_reconstructions()

        cluster_config = nabu_params.pop("cluster_config", None)
        if cluster_config == {}:
            cluster_config = None
        elif isinstance(cluster_config, SlurmClusterConfiguration):
            cluster_config = cluster_config.to_dict()

        if "tomwer_slices" in nabu_params:
            del nabu_params["tomwer_slices"]

        if "phase" in nabu_params and "delta_beta" in nabu_params["phase"]:
            pag_dbs = nabu_params["phase"]["delta_beta"]
            if isinstance(pag_dbs, str):
                pag_dbs = utils.retrieve_lst_of_value_from_str(pag_dbs, type_=float)
            if len(pag_dbs) > 1:
                raise ValueError(
                    "Several value of delta / beta found for volume reconstruction"
                )

        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(f"input type {scan} is not managed")

        if cluster_config is None:
            target = Target.LOCAL
        else:
            target = Target.SLURM

        state = None
        details = None
        try:
            self._current_processing = VolumeRunner(
                scan=scan,
                config_nabu=nabu_params,
                cluster_config=cluster_config,
                dry_run=self._dry_run,
                target=target,
                process_name=self.program_name,
            )
            res = self._current_processing.run()[0]
        except Exception as e:
            _logger.error(f"Failed to process {scan}. Error is {e}")
            state = DatasetState.FAILED
            details = None
            future_tomo_obj = None
        else:
            # tag latest reconstructions
            if isinstance(res, ResultsLocalRun) and res.results_identifiers is not None:
                scan.set_latest_vol_reconstructions(res.results_identifiers)
            # create future if needed
            if isinstance(res, ResultSlurmRun):
                future_tomo_obj = FutureTomwerObject(
                    tomo_obj=scan,
                    futures=tuple(res.future_slurm_jobs),
                    process_requester_id=self.process_id,
                )

            else:
                future_tomo_obj = None

            if self._cancelled:
                state = DatasetState.CANCELLED
                details = "cancelled by user"
                _logger.info(f"Slices computation for {scan} cancelled")
            else:
                succeed = res.success
                stdouts = (
                    [
                        res.std_out,
                    ]
                    if hasattr(res, "std_out")
                    else []
                )
                stderrs = (
                    [
                        res.std_err,
                    ]
                    if hasattr(res, "std_err")
                    else []
                )

                if not succeed:
                    mess = f"Volume computed for {scan} failed."
                    _logger.processFailed(mess)
                    state = DatasetState.FAILED
                else:
                    mess = f"Volume computed for {scan}."
                    _logger.processSucceed(mess)
                    state = DatasetState.SUCCEED

                # format stderr and stdout
                elmts = [
                    format_stderr_stdout(stderr=stderr, stdout=stdout)
                    for stderr, stdout in zip(stderrs, stdouts)
                ]
                elmts.insert(0, mess)
                details = "\n".join(elmts)
        finally:
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self,
                state=state,
                details=details,
            )

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan
        self.outputs.volumes = scan.latest_vol_reconstructions
        self.outputs.future_tomo_obj = future_tomo_obj

        # build screenshots
        if scan.latest_vol_reconstructions is not None:
            try:
                screenshots = {}
                [
                    screenshots.update(
                        select_screenshot_from_volume(
                            VolumeFactory.create_tomo_object_from_identifier(
                                rec_identifier
                            )
                        )
                    )
                    for rec_identifier in scan.latest_vol_reconstructions
                ]
            except Exception as e:
                _logger.error(f"screenshot creation failed. Error is {e}")
            else:
                self.outputs.screenshots = IcatScreenshots(
                    data_dir=deduce_dataset_gallery_location(scan),
                    screenshots=screenshots,
                    scan=scan,
                )
        else:
            self.outputs.screenshots = tuple()

    def set_configuration(self, configuration: dict) -> None:
        Task.set_configuration(self, configuration=configuration)
        if "dry_run" in configuration:
            self.set_dry_run(bool(configuration["dry_run"]))

    @staticmethod
    def program_name():
        return "nabu-volume"

    @staticmethod
    def program_version():
        return nabu_version

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @staticmethod
    def retrieve_last_relative_cor(scan):
        with EntryReader(scan.process_file_url) as h5f:
            latest_nabu_node = Task.get_most_recent_process(h5f, NabuVolumeTask)
            path = "configuration/reconstruction/rotation_axis_position"
            if latest_nabu_node is not None and path in latest_nabu_node:
                return h5py_read_dataset(latest_nabu_node[path])

    def cancel(self):
        """
        stop current processing
        """
        self._cancelled = True
        if self._current_processing is not None:
            self._current_processing.cancel()


class NabuVolume(NabuVolumeTask):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        deprecated_warning(
            name="tomwer.core.process.reconstruction.nabu.nabuvolume.NabuVolume",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="NabuVolumeTask",
        )
        super().__init__(process_id, varinfo, inputs, node_id, node_attrs, execinfo)



# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "28/08/2020"


import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt
from silx.io.url import DataUrl

from tomwer.core.process.edit import darkflatpatch
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils.lbsram import is_low_on_memory

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class DarkFlatPatchProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisProcess` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, parent=None, process_id=None):
        qt.QObject.__init__(self, parent=parent)
        FIFO.__init__(self, process_id=process_id)
        self._dry_run = False

    def _process(self, data, configuration, callback=None):
        _logger.info(f"dark-flat patch stack is processing {data}")
        self._data_currently_computed = data
        assert isinstance(data, TomwerScanBase)
        self._computationThread.finished.connect(self._end_threaded_computation)

        if isOnLbsram(data) and is_low_on_memory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = "low memory, skip dark-flat-patch reconstruction for", data.path
            _logger.processSkipped(mess)
            ProcessManager().notify_dataset_state(
                dataset=data,
                process=self,
                state=DatasetState.SKIPPED,
                details=mess,
            )
            self._end_threaded_computation()
        else:
            self._computationThread.init(data=data, configuration=configuration)
            # need to manage connect before starting it because
            ProcessManager().notify_dataset_state(
                dataset=data,
                process=self,
                state=DatasetState.ON_GOING,
                details=None,
            )
            self._computationThread.start()

    def _end_threaded_computation(self, callback=None):
        self._computationThread.finished.disconnect(self._end_threaded_computation)
        super()._end_threaded_computation(callback=callback)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _DarkFlatPatchProcessingThread(process_id=process_id)


class _DarkFlatPatchProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of dark-flat patch
    """

    def __init__(self, process_id=None):
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        SuperviseProcess.__init__(self, process_id=process_id)
        self._scan = None
        self._configuration = None

    def init(self, data, configuration):
        self._scan = data
        self._configuration = configuration

    def run(self):
        # TODO: must be replace by calling the 'DarkFlatPatch' task directly
        self.sigComputationStarted.emit()
        _logger.processStarted(f"{self._scan} Start dark-flat patch")
        process = darkflatpatch.DarkFlatPatchTask(
            inputs={
                "data": self._scan,
                "configuration": self._configuration,
                "serialize_output_data": False,
            }
        )
        try:
            process.run()
        except Exception as e:
            info = f"Fail to patch dark-flat for {self._scan}. Reason is {e}"
            _logger.processFailed(info)
            ProcessManager().notify_dataset_state(
                dataset=self._scan,
                process=self,
                state=DatasetState.FAILED,
                details=info,
            )
            return

        index = self._scan.pop_process_index()
        if isinstance(self._scan, EDFTomoScan):
            entry = None
        else:
            entry = self._scan.entry
        configuration_to_dump = self._configuration
        keys = list(configuration_to_dump.keys())
        for key in keys:
            if isinstance(configuration_to_dump[key], DataUrl):
                configuration_to_dump[key] = configuration_to_dump[key].path()

        with self._scan.acquire_process_file_lock():

            try:
                darkflatpatch.DarkFlatPatchTask._register_process(
                    process_file=self._scan.process_file,
                    entry=entry,
                    configuration=configuration_to_dump,
                    results={},
                    process=darkflatpatch.DarkFlatPatchTask,
                    process_index=index,
                    overwrite=True,
                )
            except Exception as e:
                _logger.warning(
                    f"Fail to register DarkFlatPatch process. Reason is {e}"
                )
        info = f"Dark-flat patched for {self._scan}."
        _logger.processSucceed(info)
        ProcessManager().notify_dataset_state(
            dataset=self._scan,
            process=self,
            state=DatasetState.SUCCEED,
            details=info,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import signal
import sys

from silx.gui import qt

from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefsTask
from tomwer.core.process.reconstruction.darkref.params import DKRFRP, ReduceMethod
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.reconstruction.darkref.darkrefwidget import DarkRefWidget
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.synctools.darkref import QDKRFRP

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer darkref [scanDir]"


def _exec_without_interaction(scan, dark_method, flat_method, overwrite):
    recons_params = DKRFRP()
    recons_params.overwrite_dark = overwrite
    recons_params.overwrite_ref = overwrite
    recons_params.dark_calc_method = dark_method
    recons_params.flat_calc_method = flat_method
    dark_ref = DarkRefsTask(
        inputs={
            "data": scan,
            "dark_ref_params": recons_params,
            "serialize_output_data": False,
        }
    )
    _logger.info(f"Start processing of {scan}")
    dark_ref.run()
    _logger.info(f"End processing of {scan}")
    return 0


def _exec_with_interaction(scan, dark_method, flat_method, overwrite):
    class ProcessingThread(qt.QThread):
        """
        Thread used to run the processing
        """

        def __init__(self, inputs: dict) -> None:
            super().__init__()
            self._inputs = inputs

        def run(self):
            process = DarkRefsTask(inputs=self._inputs)
            process.run()

    class _DarkRefWidgetRunnable(DarkRefWidget):
        sigScanReady = qt.Signal(str)

        def __init__(self, scan, parent=None):
            self.__scan = scan
            self.__darkref_rp = QDKRFRP()
            DarkRefWidget.__init__(self, parent=parent, reconsparams=self.__darkref_rp)
            buttonExec = qt.QPushButton("execute", parent=self)
            buttonExec.setAutoDefault(True)
            # needed to be used as an application to return end only when the
            # processing thread is needed
            self._forceSync = True
            self.layout().addWidget(buttonExec)
            buttonExec.pressed.connect(self._process)
            self.setWindowIcon(icons.getQIcon("tomwer"))
            self._processingThread = None

        def _notifyEnd(self, *args, **kwargs):
            self._processingThread.finished.disconnect(self._notifyEnd)
            print(f"computation of {self.__scan} reduced dark and flat done.")
            self._processingThread = None

        def _process(self):
            if self._processingThread is not None:
                print("processing already on-going...")
            else:
                self._processingThread = ProcessingThread(
                    inputs={
                        "data": self.__scan,
                        "dark_ref_params": self.__darkref_rp,
                        "force_sync": self._forceSync,
                    }
                )
                self._processingThread.finished.connect(self._notifyEnd)
                print(f"start reducing dark and flat of {self.__scan} ...")
                self._processingThread.start()

    def sigintHandler(*args):
        """Handler for the SIGINT signal."""
        qt.QApplication.quit()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catched
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()

    widget = _DarkRefWidgetRunnable(scan)
    # set up
    widget.recons_params.overwrite_dark = overwrite
    widget.recons_params.overwrite_flat = overwrite
    widget.recons_params.dark_calc_method = dark_method
    widget.recons_params.flat_calc_method = flat_method
    splash.finish(widget)
    widget.show()
    return app.exec_()


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path", help="Data file to show (h5 file, edf files, spec files)"
    )
    parser.add_argument(
        "--entry",
        help="an entry can be specify in case of hdf5 the master file",
        default=None,
    )
    parser.add_argument(
        "--dark-method",
        help="Define the method to be used for computing dark",
        default=ReduceMethod.MEAN,
    )
    parser.add_argument(
        "--flat-method",
        help="Define the method to be used for computing flat",
        default=ReduceMethod.MEDIAN,
    )
    parser.add_argument(
        "--no-gui",
        help="Will run directly the dark and ref without any interaction",
        dest="run",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--overwrite",
        dest="overwrite",
        action="store_true",
        default=False,
        help="Overwrite dark/flats if exists",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    increase_max_number_file()

    scan = ScanFactory.create_scan_object(options.scan_path, entry=options.entry)
    scan.set_process_index_frm_tomwer_process_file()

    dark_method = ReduceMethod.from_value(options.dark_method)
    flat_method = ReduceMethod.from_value(options.flat_method)

    if options.run:
        exit(
            _exec_without_interaction(
                scan=scan,
                dark_method=dark_method,
                flat_method=flat_method,
                overwrite=options.overwrite,
            )
        )
    else:
        exit(
            _exec_with_interaction(
                scan=scan,
                dark_method=dark_method,
                flat_method=flat_method,
                overwrite=options.overwrite,
            )
        )


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/07/2021"


import shutil
import tempfile
import unittest

import h5py
import numpy
from tomoscan.normalization import Method

from tomwer.core.process.reconstruction.normalization import normalization, params
from tomwer.core.utils.scanutils import MockNXtomo


class TestNormalization(unittest.TestCase):
    """
    Test the normalization process
    """

    def setUp(self) -> None:
        self.tempdir = tempfile.mkdtemp()
        dim = 100
        self.scan = MockNXtomo(
            scan_path=self.tempdir,
            n_proj=2,
            n_ini_proj=2,
            scan_range=180,
            dim=dim,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
        ).scan

    def tearDown(self) -> None:
        shutil.rmtree(self.tempdir)

    def testManualROI(self):
        # step1: rewrite the detector data to simplify result check
        with h5py.File(self.scan.master_file, mode="a") as h5f:
            dataset = h5f["/entry/instrument/detector/data"]
            assert dataset.shape == (2, 100, 100)  # pylint: disable=E1101
            del h5f["/entry/instrument/detector/data"]
            h5f["/entry/instrument/detector/data"] = numpy.arange(
                100 * 100 * 2
            ).reshape(2, 100, 100)

        process_params = normalization.SinoNormalizationParams()
        process_params.method = Method.SUBTRACTION
        process_params.source = params._ValueSource.MANUAL_ROI
        expected_results = {
            "mean": numpy.array([800.5, 10800.5]),
            "median": numpy.array([800.5, 10800.5]),
        }

        for calc_fct in params._ValueCalculationFct.values():
            with self.subTest(calc_fct=calc_fct):
                process_params.extra_infos = {
                    "start_x": 0,
                    "end_x": 2,
                    "start_y": 8,
                    "end_y": 9,
                    "calc_fct": calc_fct,
                    "calc_area": "volume",
                }
                process = normalization.SinoNormalizationTask(
                    inputs={
                        "data": self.scan,
                        "configuration": process_params,
                        "serialize_output_data": False,
                    }
                )
                process.run()
                res = self.scan.intensity_normalization.get_extra_infos().get("value")
                if isinstance(res, numpy.ndarray):
                    numpy.testing.assert_array_equal(res, expected_results[calc_fct])
                else:
                    numpy.testing.assert_array_equal(res, expected_results[calc_fct])



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/11/2021"


from tomwer.core.process.reconstruction.darkref.darkrefs import (
    requires_reduced_dark_and_flat,
)
from tomwer.core.utils.scanutils import MockEDF, MockNXtomo


def test_quick_run_necessary_edf(tmpdir):
    """test the `quick_run_necessary` function for EDFTomoScan"""
    scan = MockEDF.mockScan(scanID=str(tmpdir), start_dark=True, start_flat=True)
    assert scan.reduced_darks in (None, {})
    assert scan.reduced_flats in (None, {})
    requires_reduced_dark_and_flat(scan=scan)
    assert len(scan.reduced_darks) == 1
    assert len(scan.reduced_flats) == 1


def test_quick_run_necessary_hdf5(tmpdir):
    """test the `quick_run_necessary` function for NXtomoScan"""
    scan = MockNXtomo(
        scan_path=tmpdir,
        n_proj=20,
        n_ini_proj=20,
        dim=10,
    ).scan
    assert scan.reduced_darks in (None, {})
    assert scan.reduced_flats in (None, {})
    computed = requires_reduced_dark_and_flat(scan=scan)
    assert len(computed) == 2
    assert len(scan.reduced_darks) == 1
    assert len(scan.reduced_flats) == 1



from silx.gui import qt

from tomwer.gui import icons


def splash_screen():
    """

    :return: splash screen for orange-canvas
    :rtype: tuple(QPixmap, QRect),

    note: QRect is used by orange to define a mask to display the message.
          In our case we overwrite the QSplashScreen so we don't need this.
    """
    pixmap = icons.getQPixmap("tomwer")
    return pixmap, qt.QRect(0, 0, 400, 400)


def getIcon():
    """

    :return: application icon
    :rtype: QIcon
    """
    pixmap = icons.getQPixmap("tomwer")
    return qt.QIcon(pixmap)



# from orangewidget import gui
# from orangewidget.settings import Setting

# from silx.gui import qt
# import weakref
# from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThreadPerRun

# import tomwer.core.process.icat.createscreenshots
# from tomwer.gui.icat.createscreenshots import CreateRawDataScreenshotsWidget
# from processview.core.superviseprocess import SuperviseProcess


# class RawDataScreenshotCreatorOW(
#     OWEwoksWidgetOneThreadPerRun,
#     SuperviseProcess,
#     ewokstaskclass=tomwer.core.process.icat.createscreenshots.CreateRawDataScreenshotsTask,
# ):
#     """
#     This widget can receive 'data' (scan) and but some screenshot to be pushed on GALLERY.
#     """

#     name = "Create raw data screenshots"
#     id = "orangecontrib.widgets.tomwer.icat.RawDataScreenshotCreatorOW.RawDataScreenshotCreatorOW"
#     description = "Create screenshot of the raw data from user wishes. \n"
#     icon = "icons/raw_screenshots.svg"
#     priority = 63
#     keywords = [
#         "tomography",
#         "tomwer",
#         "tomo_obj",
#         "screenshot",
#         "raw data",
#     ]

#     want_main_area = True
#     want_control_area = False
#     resizing_enabled = True

#     _ewoks_default_inputs = Setting(
#         {
#             "raw_projections_required": True,
#             "raw_projections_each": 90,
#             "raw_darks_required": True,
#             "raw_flats_required": True,
#         }
#     )

#     _ewoks_inputs_to_hide_from_orange = (
#         "__process__",
#         "raw_projections_required",
#         "raw_projections_each",
#         "raw_darks_required",
#         "raw_flats_required",
#     )

#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         SuperviseProcess.__init__(self)
#         self._widget = CreateRawDataScreenshotsWidget(parent=self)

#         self._box = gui.vBox(self.mainArea, self.name)
#         layout = self._box.layout()
#         layout.addWidget(self._widget)

#         # load settings
#         self._widget.setConfiguration(self._ewoks_default_inputs)

#         # connect signal / slot
#         self._widget.sigConfigChanged.connect(self._updateSettings)

#     def _updateSettings(self):
#         self._ewoks_default_inputs = self._widget.getConfiguration()

#     def handleNewSignals(self) -> None:
#         """Invoked by the workflow signal propagation manager after all
#         signals handlers have been called.
#         """
#         # for now we want to avoid propagation any processing.
#         # task will be executed only when the user validates the dialog
#         if self.get_task_input_value("data", None) is None:
#             return
#         else:
#             super().handleNewSignals()

#     def get_task_inputs(self):
#         task_inputs = super().get_task_inputs()
#         configuration = self._widget.getConfiguration()
#         task_inputs["raw_projections_required"] = configuration[
#             "raw_projections_required"
#         ]
#         task_inputs["raw_projections_each"] = configuration["raw_projections_each"]
#         task_inputs["raw_darks_required"] = configuration["raw_darks_required"]
#         task_inputs["raw_flats_required"] = configuration["raw_flats_required"]
#         task_inputs["__process__"] = weakref.ref(self)
#         return task_inputs

#     def sizeHint(self):
#         return qt.QSize(500, 200)



import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from silx.gui import qt

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.edit.nxtomoeditor import NXtomoEditorDialog as _NXtomoEditorDialog

from ...orange.managedprocess import SuperviseOW

_logger = logging.getLogger(__name__)


class NXtomoEditorDialog(_NXtomoEditorDialog):
    def __init__(self, parent, *args, **kwargs) -> None:
        assert isinstance(parent, SuperviseOW)
        self._ow = parent
        # we need to save it become looks like orange is doing some stuff with parenting
        super().__init__(parent, *args, **kwargs)

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Ok).released.connect(
            self._overwriteNXtomo
        )

    def _overwriteNXtomo(self, *args, **kwargs):
        scan = self.mainWidget.getScan()
        if scan is not None:
            assert isinstance(self._ow, SuperviseOW)
            try:
                self.overwriteNXtomo()
            except Exception as e:
                _logger.error(
                    f"Fail to overwrite NXtomo ({scan.get_identifier().to_str()}). Error is {e}"
                )
                self._ow.notify_failed(scan=scan)
            else:
                self._ow.notify_succeed(scan=scan)
            self._ow._validateScan(scan)


class NXtomoEditorOW(SuperviseOW):
    """
    Widget to edit manually an NXtomo
    """

    name = "nxtomo-editor"
    id = "orange.widgets.tomwer.edit.NXtomoEditorOW.NXtomoEditorOW"
    description = "Interface to edit manually a NXtomo"
    icon = "icons/nx_tomo_editor.svg"
    priority = 10
    keywords = [
        "hdf5",
        "nexus",
        "tomwer",
        "file",
        "edition",
        "NXTomo",
        "editor",
        "energy",
        "distance",
        "pixel size",
    ]

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    want_main_area = True
    resizing_enabled = True

    sigScanReady = qt.Signal(str)
    # used for test only for now

    settings = Setting(dict())

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        self.widget = NXtomoEditorDialog(parent=self, hide_lockers=False)
        _layout.addWidget(self.widget)
        self.widget.mainWidget.sigEditingFinished.connect(self._updateSettings)
        # load settings
        self.widget.setConfiguration(self.settings)

    def _updateSettings(self):
        self.settings = self.getConfiguration()

    def _validateScan(self, scan):
        self.Outputs.data.send(scan)
        self.sigScanReady.emit(str(scan))
        super().hide()

    @Inputs.data
    def setScan(self, scan):
        self._setScan(scan=scan)

    def _setScan(self, scan):
        if scan is None:
            pass
        elif not isinstance(scan, NXtomoScan):
            raise TypeError(
                f"expect to have an instance of {NXtomoScan}. {type(scan)} provided."
            )
        else:
            self.widget.setScan(scan)
            if self.widget.hasLockField():
                self.widget._overwriteNXtomo()
                # note: _overwriteNXtomo will trigger the scan validation
            else:
                self.show()
                self.raise_()

    def sizeHint(self):
        return qt.QSize(400, 500)

    # expose API
    def getConfiguration(self) -> dict:
        return self.widget.getConfiguration()

    def setConfiguration(self, config: dict):
        self.widget.setConfiguration(config)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "28/10/2020"


import logging

from orangewidget import gui

import tomwer.core.process.edit.imagekeyeditor
from tomwer.gui.edit.imagekeyeditor import ImageKeyUpgraderWidget

from ...orange.managedprocess import TomwerWithStackStack

_logger = logging.getLogger(__name__)


class ImageKeyUpgraderOW(
    TomwerWithStackStack,
    ewokstaskclass=tomwer.core.process.edit.imagekeyeditor.ImageKeyUpgraderTask,
):
    """
    Widget to define upgrade all frames with some specific values of 'image_key' others values
    """

    name = "image-key-upgrader"
    id = "orange.widgets.tomwer.control.ImageKeyUpgraderOW.ImageKeyUpgraderOW"
    description = "Interface to upgrade all values of one or several specific `image_key` to a different value"
    icon = "icons/image_key_upgrader.svg"
    priority = 60
    keywords = [
        "hdf5",
        "nexus",
        "tomwer",
        "file",
        "edition",
        "NXTomo",
        "editor",
        "image key upgrader",
        "image-key-upgrader",
        "image_key",
        "image_key_control",
    ]

    want_main_area = True
    resizing_enabled = True

    _ewoks_inputs_to_hide_from_orange = ("operations", "serialize_output_data")

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        self.widget = ImageKeyUpgraderWidget(parent=self)
        _layout.addWidget(self.widget)

        # connect signal / slot
        if isinstance(self.task_output_changed_callbacks, set):
            self.task_output_changed_callbacks.add(self._notify_state)
        elif isinstance(self.task_output_changed_callbacks, list):
            self.task_output_changed_callbacks.append(self._notify_state)
        else:
            raise NotImplementedError
        self.widget.sigOperationsChanged.connect(self._update_operations_to_ewoks)

        self._updateGUIFromDefaultValues()
        # in case there is information registered and the GUI is not synchronized with the default values
        self._update_operations_to_ewoks()

    def _update_operations_to_ewoks(self):
        self.update_default_inputs(operations=self.widget.getOperations())

    def _updateGUIFromDefaultValues(self):
        try:
            operations = self.get_default_input_values()["operations"]
        except Exception:
            operations = {}
        self.widget.setOperations(operations)

    def get_task_inputs(self):
        task_inputs = super().get_task_inputs()
        try:
            scan = task_inputs["data"]
        except Exception:
            pass
        else:
            self.notify_pending(scan=scan)
        return task_inputs

    def _execute_ewoks_task(self, propagate):
        task_arguments = self._get_task_arguments()
        scan = task_arguments.get("inputs", {}).get("data", None)
        if scan is not None:
            self.notify_pending(scan=scan)
        super()._execute_ewoks_task(propagate=propagate)

    def _notify_state(self):
        try:
            task_executor = self.sender()
            task_suceeded = task_executor.succeeded
            scan = task_executor.current_task.inputs.data
            if task_suceeded:
                self.notify_succeed(scan=scan)
            else:
                self.notify_failed(scan=scan)
        except Exception as e:
            _logger.error(f"failed to handle task finished callback. Raiseon is {e}")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""
module dealing with log in color

https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
"""

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "22/09/2017"

import logging
import os
from sys import platform

from tomwer.core.log.processlog import (
    PROCESS_ENDED_NAME,
    PROCESS_FAILED_NAME,
    PROCESS_INFORM_NAME,
    PROCESS_SKIPPED_NAME,
    PROCESS_STARTED_NAME,
    PROCESS_SUCCEED_NAME,
)

# The background is set with 40 plus the number of the color, and the
# foreground with 30

# These are the sequences need to get colored ouput
_RESET_SEQ = "\033[0m"
_BOLD_SEQ = "\033[1m"

_BLACK = "\033[30m"
_RED = "\033[31m"
_GREEN = "\033[32m"
_YELLOW = "\033[33m"
_BLUE = "\033[34m"
_MAGENTA = "\033[35m"

LOG_COLORS = {
    "WARNING": _MAGENTA,
    "INFO": _BLACK,
    "DEBUG": _BLUE,
    "CRITICAL": _YELLOW,
    "ERROR": _RED,
    PROCESS_SKIPPED_NAME: _MAGENTA,
    PROCESS_ENDED_NAME: _BLACK,
    PROCESS_INFORM_NAME: _BLACK,
    PROCESS_STARTED_NAME: _BLACK,
    PROCESS_FAILED_NAME: _RED,
    PROCESS_SUCCEED_NAME: _GREEN,
}


def _formatter_message(message, use_color=True):
    if use_color is True:
        message = message.replace("$RESET", _RESET_SEQ).replace("$BOLD", _BOLD_SEQ)
    else:
        message = message.replace("$RESET", "").replace("$BOLD", "")
    return message


class _ColoredFormatter(logging.Formatter):
    def __init__(self, msg):
        logging.Formatter.__init__(self, msg)
        if platform == "win32":
            self.use_color = False
        else:
            self.use_color = os.environ.get("ORANGE_COLOR_STDOUT_LOG", "True") == "True"

    def format(self, record):
        levelname = record.levelname
        if self.use_color is True and levelname in LOG_COLORS:
            levelname_color = LOG_COLORS[levelname] + levelname + _RESET_SEQ
            record.levelname = levelname_color
        record.asctime = self.formatTime(record, self.datefmt)
        return logging.Formatter.format(self, record)


class TomwerLogger(logging.Logger):
    """
    Custom logger class with multiple destinations
    """

    FORMAT = (
        "%(asctime)s [%(levelname)-16s] %(message)s [$BOLD%(name)-20s$RESET]"
        "($BOLD%(filename)s$RESET:%(lineno)d)"
    )

    def __init__(self, name):
        logging.Logger.__init__(self, name, logging.WARNING)
        self.name = name
        self.color_format = _formatter_message(
            TomwerLogger.FORMAT,
            use_color=os.environ.get("ORANGE_COLOR_STDOUT_LOG", "False") == "True",
        )

        color_formatter = _ColoredFormatter(self.color_format)

        console = logging.StreamHandler()
        console.setFormatter(color_formatter)

        self.addHandler(console)
        return

    def __str__(self):
        return self.name



import functools
import numpy
from silx.gui import qt

from tomwer.gui.stitching.axisorderedlist import AxisOrderedTomoObjsModel
from tomwer.gui.utils.step import StepSizeSelectorWidget


class PosEditorOverOneAxis(qt.QWidget):
    """keep it ordered along one axis"""

    DEFAULT_VALUE_WHEM_MISSING = 0

    def __init__(
        self, parent, axis_edited: int, axis_order=None, *args, **kwargs
    ) -> None:
        assert axis_edited in (0, 1, 2)
        super().__init__(parent, *args, **kwargs)
        self._axisEdited = axis_edited
        # the axis the spin boxes are editing
        self._axisOrder = axis_order if axis_order is not None else axis_edited
        # the axis along which the tomo obj are ordered
        self._tomoObjtoSpinBoxes = {}
        # list of spin box to edit the position over the axis. Key is the tomo object, value is the spin box
        self.__spinBoxescallback = {}
        self.setLayout(qt.QVBoxLayout())
        # widget to define step size
        self._stepSizeWidget = StepSizeSelectorWidget(
            self,
            fine_value=1,
            medium_value=5,
            rough_value=25,
            dtype=int,
        )
        self.layout().addWidget(self._stepSizeWidget)

        # table with the different Tomo objects
        self._tomoObjsTableView = qt.QTableView(parent=self)
        model = EditableAxisOrderedTomoObjsModel(axis=self._axisOrder)
        self._tomoObjsTableView.setModel(model)

        self.layout().addWidget(self._tomoObjsTableView)

        # connect signal / slot
        self._stepSizeWidget.valueChanged.connect(self._updateStepSize)

        # tune table view
        self._tomoObjsTableView.setColumnWidth(0, 15)
        self._tomoObjsTableView.setColumnWidth(2, 120)
        self._tomoObjsTableView.horizontalHeader().setSectionResizeMode(
            1, qt.QHeaderView.Stretch
        )
        self.setStepSize(1)

    def _updateStepSize(self):
        step_size = self.getStepSize()
        for sb in self._tomoObjtoSpinBoxes.values():
            sb.setSingleStep(step_size)

    def setStepSize(self, step_size: int):
        self._stepSizeWidget.setStepSize(step_size)
        self._updateStepSize()

    def getStepSize(self) -> int:
        return self._stepSizeWidget.getStepSize()

    def addTomoObj(self, tomo_obj):
        if tomo_obj is None:
            return
        else:
            self._tomoObjsTableView.model().addTomoObj(tomo_obj)
            # register tomo obj metadata modification to make sure we keel the z ordered list up to data
            tomo_obj.stitching_metadata.sigChanged.connect(
                self._orderedMightHavechanged
            )
            self._createSpinBox(tomo_obj=tomo_obj)
            self._orderedMightHavechanged()

    def _createSpinBox(self, tomo_obj):
        spinBox = qt.QSpinBox(parent=self)
        spinBox.setRange(numpy.iinfo(numpy.int32).min, numpy.iinfo(numpy.int32).max)
        spinBox.setSuffix("px")
        spinBox.setSingleStep(self.getStepSize())
        spinBox.setValue(
            tomo_obj.stitching_metadata.get_abs_position_px(axis=self._axisEdited)
            or self.DEFAULT_VALUE_WHEM_MISSING
        )
        identifier_as_str = tomo_obj.get_identifier().to_str()
        self._tomoObjtoSpinBoxes[identifier_as_str] = spinBox
        # connect signal / slot
        callback = functools.partial(
            self._spinBoxValueChanged, spin_box=spinBox, tomo_obj=tomo_obj
        )
        self.__spinBoxescallback[identifier_as_str] = callback
        spinBox.editingFinished.connect(callback)
        return spinBox

    def _deleteSpinBox(self, tomo_obj):
        identifier_as_str = tomo_obj.get_identifier().to_str()
        if identifier_as_str in self._tomoObjtoSpinBoxes:
            del self._tomoObjtoSpinBoxes[identifier_as_str]
            del self.__spinBoxescallback[identifier_as_str]

    def removeTomoObj(self, tomo_obj):
        self._deleteSpinBox(tomo_obj)
        self._tomoObjsTableView.model().removeTomoObj(tomo_obj)
        tomo_obj.stitching_metadata.sigChanged.disconnect(self._orderedMightHavechanged)

    def _spinBoxValueChanged(self, spin_box, tomo_obj, *args, **kwargs):
        tomo_obj.stitching_metadata.setPxPos(spin_box.value(), axis=self._axisEdited)

    def setTomoObjs(self, tomo_objs: tuple) -> None:
        """
        replace current list of object by the given list
        """
        self._tomoObjsTableView.model().clearTomoObjs()
        for tomo_obj in tomo_objs:
            self.addTomoObj(tomo_obj)

    def clean(self):
        tomo_objs = self._tomoObjsTableView.model().getTomoObjs()
        for tomo_obj in tomo_objs:
            self.removeTomoObj(tomo_obj=tomo_obj)

    def reorderAlongAxis(self):
        pass

    def setSerie(self):
        raise NotImplementedError

    def _orderedMightHavechanged(self, force_sb_update=False):
        # add index widget
        self._tomoObjsTableView.model().reorder_objs()
        self._tomoObjsTableView.model().layoutChanged.emit()

        ordered_objs = self._tomoObjsTableView.model()._axis_decreasing_ordered_objs

        # start update spin box positions
        # check if we need to update one widget. Has this is designed if we need to update one then we need to update them all...
        needs_to_update_widget = force_sb_update
        for i_pos, tomo_obj in enumerate(ordered_objs):
            identifier_as_str = tomo_obj.get_identifier().to_str()
            spinBox = self._tomoObjtoSpinBoxes[identifier_as_str]
            model_index = self._tomoObjsTableView.model().createIndex(i_pos, 2)
            needs_to_update_widget = (
                needs_to_update_widget
                or self._tomoObjsTableView.indexWidget(model_index)
                not in (None, spinBox)
            )

        for i_pos, tomo_obj in enumerate(ordered_objs):
            identifier_as_str = tomo_obj.get_identifier().to_str()
            spinBox = self._tomoObjtoSpinBoxes[identifier_as_str]
            model_index = self._tomoObjsTableView.model().createIndex(i_pos, 2)
            if needs_to_update_widget:
                # if items have been reordered then we must recreated SpinBoxes otherwise if we try
                # to change order then Qt will end up with a seg fault which seems to come from
                # otherwirtting the cell and trying to reuse them
                self._deleteSpinBox(tomo_obj=tomo_obj)
                spinBox = self._createSpinBox(tomo_obj=tomo_obj)
            self._tomoObjsTableView.setIndexWidget(model_index, spinBox)
        # end update spin box positions


class EditableAxisOrderedTomoObjsModel(AxisOrderedTomoObjsModel):
    def __init__(self, axis: int, parent=None) -> None:
        super().__init__(axis, parent)
        self._headers = ["index", "tomo obj", f"axis {axis} pos (px)"]



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "19/07/2018"


from silx.gui import qt

from tomwer.core.process.conditions.filters import RegularExpressionFilter


class RegularExpressionSandBoxDialog(qt.QDialog):
    """
    Dialog for showing the NameFilterSandBoxDialog
    """

    def __init__(self, parent=None, pattern=None):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self.setWindowTitle("Regular expression sand box")

        self.sandbox = RegularExpressionSandBox(parent=self, pattern=pattern)
        self.layout().addWidget(self.sandbox)

        types = qt.QDialogButtonBox.Close
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)
        self._buttons.button(qt.QDialogButtonBox.Close).clicked.connect(self.hide)

        # expose API
        self.setPattern = self.sandbox.setPattern

        self.setPattern(pattern=pattern)


class RegularExpressionSandBox(RegularExpressionFilter, qt.QWidget):
    """
    Widget to see if the pattern will 'valid' the given value or not
    """

    def __init__(self, parent=None, pattern=None):
        _pattern = pattern
        if _pattern is None:
            _pattern = ""
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._docLabel = qt.QLabel(
            '<a href="https://docs.python.org/3/library/re.html"> See documentation</a>',
            self,
        )
        self._docLabel.setOpenExternalLinks(True)
        self.layout().addWidget(self._docLabel)

        self._valueLE = qt.QLineEdit("", parent=self)
        self._valueLE.editingFinished.connect(self.applyFilter)
        self._valueLE.setAlignment(qt.Qt.AlignCenter)
        self.layout().addWidget(self._valueLE)

        style = qt.QApplication.style()

        # down arrow
        self._arrowWidget = qt.QWidget(parent=self)
        self._arrowWidget.setLayout(qt.QHBoxLayout())
        self._arrowWidget.setContentsMargins(0, 0, 0, 0)
        self._arrowWidget.layout().setSpacing(0)
        self._arrow = qt.QLabel(parent=self._arrowWidget)
        self._arrow.setMinimumSize(qt.QSize(55, 55))
        icon = style.standardIcon(qt.QStyle.SP_ArrowDown)
        self._arrow.setPixmap(icon.pixmap(55, 55))
        slicer1 = qt.QWidget(parent=self._arrowWidget)
        slicer1.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        slicer2 = qt.QWidget(parent=self._arrowWidget)
        slicer2.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._arrowWidget.layout().addWidget(slicer1)
        self._arrowWidget.layout().addWidget(self._arrow)
        self._arrowWidget.layout().addWidget(slicer2)
        self.layout().addWidget(self._arrowWidget)

        # pattern Line edit
        self._patternWidget = qt.QWidget(parent=self)
        self._patternWidget.setLayout(qt.QHBoxLayout())
        slicer1 = qt.QWidget(parent=self._arrowWidget)
        slicer1.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        slicer2 = qt.QWidget(parent=self._arrowWidget)
        slicer2.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._patternWidget.layout().addWidget(slicer1)
        self._patternWidget.layout().addWidget(qt.QLabel("pattern:"))
        self._patternLE = qt.QLineEdit(_pattern or "", parent=self._patternWidget)
        self._patternWidget.layout().addWidget(self._patternLE)
        self._patternWidget.layout().addWidget(slicer2)
        self.layout().addWidget(self._patternWidget)
        self._patternLE.editingFinished.connect(self._updatePattern)

        # result label
        self._resultWidget = qt.QWidget(parent=self)
        self._resultWidget.setLayout(qt.QHBoxLayout())
        self._resultWidget.setContentsMargins(0, 0, 0, 0)
        self._resultWidget.layout().setSpacing(0)

        self._resultLabel = qt.QLabel(parent=self._resultWidget)
        self._resultLabel.setMinimumSize(qt.QSize(40, 40))
        slicer1 = qt.QWidget(parent=self._resultWidget)
        slicer1.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        slicer2 = qt.QWidget(parent=self._resultWidget)
        slicer2.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._resultWidget.layout().addWidget(slicer1)
        self._resultWidget.layout().addWidget(self._resultLabel)
        self._resultWidget.layout().addWidget(slicer2)
        self.layout().addWidget(self._resultWidget)
        RegularExpressionFilter.__init__(self, _pattern)

    def setPattern(self, pattern):
        self._patternLE.setText(pattern or "")
        RegularExpressionFilter.setPattern(self, pattern)
        self.applyFilter()

    def applyFilter(self):
        """
        Update result label depending on the value to check and the filter
        """
        value = self._valueLE.text()

        style = qt.QApplication.style()
        if self.isFiltered(value):
            icon = style.standardIcon(qt.QStyle.SP_DialogNoButton)
        else:
            icon = style.standardIcon(qt.QStyle.SP_DialogYesButton)
        self._resultLabel.setPixmap(icon.pixmap(40, 40))

    def _updatePattern(self):
        """callback when the pattern is modify"""
        self.blockSignals(True)
        self.setPattern(self._patternLE.text())
        self.blockSignals(False)

    @staticmethod
    def description():
        return (
            "A simple sand box to play with pattern definition and input "
            "names to make sure it will have the requested behavior."
        )

    def unvalidPatternDefinition(self, pattern, error):
        """Overwrite NameFilter.unvalidPatternDefinition"""
        txt = 'regular expression "%s" is invalid. Error is:' % pattern
        mess = qt.QMessageBox(parent=self, icon=qt.QMessageBox.warning, text=txt)
        mess.setInformativeText(str(error))
        mess.setWindowTitle("Regular expression failed")
        mess.setModal(False)
        mess.show()



import logging

from ewokscore.missing_data import is_missing_data

from sluurp.executor import submit as submit_to_slurm_cluster

from nabu.stitching.config import (
    dict_to_config_obj,
    PreProcessedZStitchingConfiguration,
    PostProcessedZStitchingConfiguration,
    SlurmConfig,
    StitchingConfiguration,
    StitchingType,
)
from nabu.stitching.z_stitching import (
    z_stitching,
    StitchingPostProcAggregation as _StitchingPostProcAggregation,
)
from nabu.stitching.slurm_utils import split_stitching_configuration_to_slurm_job

from processview.core.superviseprocess import SuperviseProcess

from tomwer.core.process.task import Task
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.volume.volumefactory import VolumeFactory


_logger = logging.getLogger(__name__)


class StitcherTask(
    Task,
    SuperviseProcess,
    input_names=("stitching_config",),
    optional_input_names=(
        "cluster_config",
        "progress",
        "serialize_output_data",
    ),
    output_names=(
        "data",
        "future_tomo_obj",
        "volume",
    ),
):
    def __init__(self, process_id=None, *args, **kwargs):
        SuperviseProcess.__init__(self, process_id=process_id)
        super().__init__(*args, **kwargs)

    def run(self):
        cluster_config = self.inputs.cluster_config
        assert cluster_config is None or isinstance(cluster_config, dict)
        stitching_config = self.inputs.stitching_config
        assert isinstance(stitching_config, dict)

        config = dict_to_config_obj(stitching_config)
        config.slurm_config = SlurmConfig.from_dict(cluster_config)
        if cluster_config in ({}, None):
            # opt1: run locally
            config.slurm_config = SlurmConfig.from_dict(cluster_config)

            progress = (
                None if is_missing_data(self.inputs.progress) else self.inputs.progress
            )
            stitched_identifier = z_stitching(configuration=config, progress=progress)
            if config.stitching_type is StitchingType.Z_PREPROC:
                nx_tomo = ScanFactory.create_tomo_object_from_identifier(
                    stitched_identifier.to_str()
                )
                if self.get_input_value("serialize_output_data", True):
                    self.outputs.data = nx_tomo.to_dict()
                else:
                    self.outputs.data = nx_tomo
                self.outputs.volume = None
                self.outputs.future_tomo_obj = None
            elif config.stitching_type is StitchingType.Z_POSTPROC:
                volume = VolumeFactory.create_tomo_object_from_identifier(
                    stitched_identifier.to_str()
                )
                self.outputs.volume = volume
                self.outputs.data = None
                self.outputs.future_tomo_obj = None
            else:
                raise NotImplementedError()
        else:
            # opt2: run remotly and aggregate locally
            futures = {}

            # 2.1 launch jobs
            for i_job, (job, sub_config) in enumerate(
                split_stitching_configuration_to_slurm_job(
                    config, yield_configuration=True
                )
            ):
                _logger.info(f"submit job nb {i_job}: handles {sub_config.slices}")
                output_tomo_obj = (
                    sub_config.get_output_object().get_identifier().to_str()
                )
                futures[output_tomo_obj] = submit_to_slurm_cluster(job, timeout=999999)

            # handle post processing
            data_aggregation = StitchingPostProcAggregation(
                futures=futures,
                stitching_config=config,
            )
            for future in futures.values():
                # TODO: do we ned to make sure the processing is not finished yet ?
                future.add_done_callback(data_aggregation.onePartFinished)

            if config.stitching_type is StitchingType.Z_PREPROC:
                output_nx_tomo_file = config.output_file_path
                output_nx_tomo_entry = config.output_data_path
                tomwer_scan = NXtomoScan(
                    scan=output_nx_tomo_file, entry=output_nx_tomo_entry
                )
                self.outputs.future_tomo_obj = FutureTomwerObject(
                    tomo_obj=tomwer_scan,
                    futures=tuple(futures.values()),
                )
                self.outputs.data = None
                self.outputs.volume = None

            elif config.stitching_type is StitchingType.Z_POSTPROC:
                tomwer_volume = VolumeFactory.create_tomo_object_from_identifier(
                    config.output_volume.get_identifier().to_str()
                )
                self.outputs.future_tomo_obj = FutureTomwerObject(
                    tomo_obj=tomwer_volume,
                    futures=tuple(futures.values()),
                )
                self.outputs.volume = None
                self.outputs.data = None


class StitchingPostProcAggregation(_StitchingPostProcAggregation):
    """
    once stitching is done remotly we need to 'aggregate' the different part in the correct order
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._n_finished = 0

        if isinstance(self.stitching_config, dict):
            stitching_type = StitchingType.from_value(
                self.stitching_config["stitching"]["type"]
            )
            if stitching_type is StitchingType.Z_PREPROC:
                self._stitching_config = PreProcessedZStitchingConfiguration.from_dict(
                    self.stitching_config
                )
            elif stitching_type is StitchingType.Z_POSTPROC:
                self._stitching_config = PostProcessedZStitchingConfiguration.from_dict(
                    self.stitching_config
                )
            else:
                raise NotImplementedError("stitching type not handled")
        elif not isinstance(self.stitching_config, StitchingConfiguration):
            raise TypeError(
                f"stitching_config is expected to be an instance of {StitchingConfiguration}. {type(self.stitching_config)} provided instead"
            )

    def onePartFinished(self, *args, **kwargs):
        self._n_finished += 1
        # note: for now we only consider the user of the futures.
        # if users want to only run post-processing agregation then they will use the nabu CLI for it.
        if self._n_finished == len(self._futures):
            _logger.info("All slurm job finished. Will aggregate results")
            self.process()



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain utils for score process
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "27/10/2021"


import logging
import os
import subprocess
import uuid
from typing import Iterable, Optional, Union
import signal
import psutil
import sys

import numpy
from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5
from sluurp.executor import submit as submit_to_slurm_cluster
from sluurp.job import SBatchScriptJob
from tomoscan.io import HDF5File
from tomoscan.normalization import Method as INormMethod
from tomoscan.identifier import VolumeIdentifier

from tomwer.core.process.reconstruction.nabu.plane import NabuPlane
from tomwer.core.cluster.cluster import SlurmClusterConfiguration
from tomwer.core.process.reconstruction.nabu.target import Target
from tomwer.core.process.reconstruction.nabu.utils import (
    _NabuPhaseMethod,
    nabu_std_err_has_error,
)
from tomwer.core.process.reconstruction.output import (
    ProcessDataOutputDirMode,
    get_output_folder_from_scan,
    NabuOutputFileFormat,
)
from tomwer.core.process.reconstruction.normalization.params import (
    _ValueSource as INormSource,
)
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.slurm import get_slurm_script_name, is_slurm_available
from tomwer.core.volume.volumefactory import VolumeFactory

from . import settings, utils

_logger = logging.getLogger(__name__)
try:
    from nabu.pipeline.fullfield.reconstruction import (  # noqa F401
        FullFieldReconstructor,
    )
except (ImportError, OSError) as e:
    try:
        from nabu.pipeline.fullfield.local_reconstruction import (  # noqa F401
            ChunkedReconstructor,
        )
    except (ImportError, OSError):
        # import of cufft library can bring an OSError if cuda not install
        _logger.error(e)
        has_nabu = False
    else:
        has_nabu = True
else:
    has_nabu = True


class ResultsRun:
    """
    Base class of results for nabu
    """

    def __init__(self, success, config) -> None:
        self.__success = success
        self.__config = config

    @property
    def success(self) -> bool:
        return self.__success

    @property
    def config(self) -> dict:
        return self.__config

    def __str__(self) -> str:
        return f"result from nabu run: {'succeed' if self.success else 'failed'} with \n - config:{self.config} \n"


class ResultsWithStd(ResultsRun):
    """Nabu result with std"""

    def __init__(self, std_out, std_err, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__std_err = std_err
        self.__std_out = std_out

    @property
    def std_out(self) -> str:
        return self.__std_out

    @property
    def std_err(self) -> str:
        return self.__std_err

    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n {self.std_out} \n {self.std_err}"
        return res


class ResultsLocalRun(ResultsWithStd):
    """Nabu result when run locally.
    If this is the case we should be able to retrieve directly the results urls"""

    def __init__(
        self,
        results_identifiers: tuple,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        if not isinstance(results_identifiers, Iterable):
            raise TypeError(
                f"results_urls is expected to be an Iterable not {type(results_identifiers)}"
            )

        # check all identifiers
        def check_identifier(identifier):
            if isinstance(identifier, str):
                vol = VolumeFactory.create_tomo_object_from_identifier(
                    identifier=identifier
                )
                return vol.get_identifier()
            elif not isinstance(identifier, VolumeIdentifier):
                raise TypeError(
                    f"identifiers are expected to be VolumeIdentifier. Get {type(identifier)} instead."
                )
            else:
                return identifier

        self.__results_identifiers = tuple(
            [
                check_identifier(identifier=identifier)
                for identifier in results_identifiers
            ]
        )

    @property
    def results_identifiers(self) -> tuple:
        return self.__results_identifiers

    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n - result urls: {self.results_identifiers}"
        return res


class ResultSlurmRun(ResultsWithStd):
    """Nabu result when run on slurm. on this case we expect to get a future and a distributed client"""

    def __init__(
        self,
        future_slurm_jobs: tuple,
        job_id: Optional[int],
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.__future_slurm_jobs = future_slurm_jobs
        self.__job_id = job_id

    @property
    def future_slurm_jobs(self):
        return self.__future_slurm_jobs

    @property
    def job_id(self) -> Optional[int]:
        return self.__job_id

    def __str__(self) -> str:
        res = super().__str__()
        res += f"\n - future job slurms: {self.future_slurm_jobs} \n"
        return res


class _NabuBaseReconstructor:
    """
    Base class to submit a job to nabu
    """

    TIMEOUT_SLURM_JOB_SUBMISSION = 30
    """Timeout when submit a job to slurm cluster. In second"""

    EXPECTS_SINGLE_SLICE = True

    def __init__(
        self,
        scan: TomwerScanBase,
        dry_run: bool,
        target: Target,
        cluster_config: Optional[Union[dict, SlurmClusterConfiguration]],
        process_name: str,
        axis: NabuPlane = NabuPlane.XY,
    ) -> None:
        self._scan = scan
        self._target = Target.from_value(target)
        self._dry_run = dry_run
        self._process_name = process_name
        self._process = None
        self._cancelled = False
        self._axis = NabuPlane.from_value(axis)
        # nabu subprocess if run locally
        if isinstance(cluster_config, SlurmClusterConfiguration):
            self._cluster_config = cluster_config
        elif isinstance(cluster_config, dict):
            self._cluster_config = SlurmClusterConfiguration.from_dict(cluster_config)
        elif cluster_config is None:
            self._cluster_config = None
        else:
            raise TypeError(
                f"cluster config is expected to be a dict or an instance of {SlurmClusterConfiguration}. Not {type(cluster_config)}"
            )

    @property
    def scan(self):
        return self._scan

    @property
    def target(self):
        return self._target

    @property
    def cluster_config(self):
        return self._cluster_config

    @property
    def dry_run(self):
        return self._dry_run

    @property
    def process_name(self):
        return self._process_name

    @property
    def axis(self) -> NabuPlane:
        return self._axis

    def only_create_config_file(self):
        """Should we run the reconstruction or only create the configuration file"""
        return False

    def run(self) -> Iterable:
        """
        run the requested slices.

        :return: Iterable of ResultsRun.
        """
        raise NotImplementedError("Base class")

    def _process_config(
        self,
        config_to_dump: dict,
        config_file: str,
        file_format: str,
        info: Optional[str],
        process_name: str,
    ):
        """
        process provided configuration

        :param str info:
        """
        if self.dry_run is True or self.only_create_config_file():
            return ResultsRun(
                success=True,
                config=config_to_dump,
            )
        elif self.target is Target.LOCAL:
            _logger.info(f"run {info} for {self.scan} with {config_to_dump}")
            return self._run_nabu_locally(
                conf_file=config_file,
                file_format=file_format,
                config_to_dump=config_to_dump,
                axis=config_to_dump["reconstruction"].get("slice_plane", "XY"),
            )
        elif self.target is Target.SLURM:
            _logger.info(
                f"run {info} on slurm for {self.scan.path} with {config_to_dump}"
            )
            return self._run_nabu_on_slurm(
                conf_file=config_file,
                config_to_dump=config_to_dump,
                cluster_config=self.cluster_config.to_dict(),
                process_name=process_name,
                info=info,
            )
        else:
            raise ValueError(f"{self.target} is not recognized as a valid target")

    @staticmethod
    def _get_gpu_and_cpu_mem_fraction(config_to_dump: dict):
        gpu_mem_fraction = config_to_dump.get("resources", {}).get(
            "gpu_mem_fraction", config_to_dump.get("gpu_mem_fraction", 0.9)
        )
        assert gpu_mem_fraction <= 1
        cpu_mem_fraction = config_to_dump.get("resources", {}).get(
            "cpu_mem_fraction", config_to_dump.get("cpu_mem_fraction", 0.9)
        )
        assert cpu_mem_fraction <= 1
        return gpu_mem_fraction, cpu_mem_fraction

    def _run_nabu_locally(
        self,
        axis: NabuPlane,
        conf_file: str,
        file_format: str,
        config_to_dump: dict,
    ) -> ResultsLocalRun:
        """
        run locally nabu for a single configuration file.

        :param str conf_file: path to the nabu .cfg file
        :param str file_format: format of the generated file
        :param dict config_to_dump: configuration saved in the .cfg as a dictionary
        :return: results of the local run
        :rtype: ResultsLocalRun
        """
        if not has_nabu:
            raise ImportError("Fail to import nabu")
        assert isinstance(config_to_dump, dict)
        gpu_mem_fraction, cpu_mem_fraction = self._get_gpu_and_cpu_mem_fraction(
            config_to_dump
        )

        command = " ".join(
            (
                sys.executable,
                "-m",
                settings.NABU_FULL_FIELD_APP_PATH,
                conf_file,
                "--gpu_mem_fraction",
                str(gpu_mem_fraction),
                "--cpu_mem_fraction",
                str(cpu_mem_fraction),
            )
        )
        _logger.info(f'call nabu from "{command}"')

        self._process = subprocess.Popen(
            command,
            shell=True,
            cwd=self.scan.path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            outs, errs = self._process.communicate()
        except (TimeoutError, KeyboardInterrupt):
            self._process.kill()
            outs, errs = self._process.communicate()

        recons_urls = utils.get_recons_volume_identifier(
            file_prefix=config_to_dump["output"]["file_prefix"],
            location=config_to_dump["output"]["location"],
            slice_index=None,
            scan=self.scan,
            file_format=file_format,
            axis=axis,
        )
        return ResultsLocalRun(
            success=not nabu_std_err_has_error(errs),
            results_identifiers=recons_urls,
            std_out=outs,
            std_err=errs,
            config=config_to_dump,  # config_slices,
        )

    def _run_nabu_on_slurm(
        self,
        conf_file: str,
        config_to_dump: dict,
        cluster_config: dict,
        process_name: str,
        info: str,
    ) -> ResultSlurmRun:
        """
        Run a nabu reconstruction on slurm of a single configuration

        :return: results of the slurm run
        :rtype: ResultSlurmRun
        """
        if not isinstance(conf_file, str):
            raise TypeError(f"conf_file is expected to be a strg not {type(conf_file)}")
        if not isinstance(config_to_dump, dict):
            raise TypeError(
                f"config_to_dump is expected to be a strg not {type(config_to_dump)}"
            )
        if not is_slurm_available():
            raise RuntimeError("slurm not available")
        if not isinstance(cluster_config, dict):
            raise ValueError(
                f"cluster config is expected to be a dict not {type(cluster_config)}"
            )

        # create slurm cluster
        project_name = cluster_config.get(
            "job_name", "tomwer_{scan}_-_{process}_-_{info}"
        )
        project_name = project_name.format(
            scan=str(self.scan), process=process_name, info=info
        )
        # project name should not contain any spaces as it will be integrated in a script and interpreted.
        project_name = project_name.replace(" ", "_")
        cluster_config["job_name"] = project_name

        # extract gpu_mem_fraction and cpu_mem_fraction
        assert isinstance(config_to_dump, dict)
        gpu_mem_fraction, cpu_mem_fraction = self._get_gpu_and_cpu_mem_fraction(
            config_to_dump
        )

        # submit job
        script_name = get_slurm_script_name(prefix="nabu")
        # for now force job name
        cluster_config["job_name"] = f"tomwer-nabu {conf_file}"
        job = SBatchScriptJob(
            slurm_config=cluster_config,
            script=(
                f"python3 -m {settings.NABU_FULL_FIELD_APP_PATH} {conf_file} --gpu_mem_fraction {gpu_mem_fraction} --cpu_mem_fraction {cpu_mem_fraction}",
            ),
            script_path=os.path.join(self.scan.path, "slurm_scripts", script_name),
            clean_script=False,
            working_directory=self.scan.working_directory,
        )
        future_slurm_job = submit_to_slurm_cluster(job)

        callbacks = self._get_futures_slurm_callback(config_to_dump)
        assert isinstance(
            callbacks, tuple
        ), f"callbacks is expected to an instance of tuple and not {type(callbacks)}"
        for callback in callbacks:
            future_slurm_job.add_done_callback(callback.process)

        return ResultSlurmRun(
            success=True,
            config=config_to_dump,
            future_slurm_jobs=(future_slurm_job,),
            std_out=None,
            std_err=None,
            job_id=job.job_id,
        )

    def _get_futures_slurm_callback(self, config_to_dump) -> tuple:
        """Return a tuple a potential callback to be launch once the future is done"""
        return tuple()

    def _treateOutputSliceConfig(self, config) -> tuple:
        """
        - add or overwrite some parameters of the dictionary
        - create the output directory if does not exist

        :return: (config: dict, nabu_cfg_folder: str)
        """
        # handle phase
        pag = False
        ctf = False
        db = None
        if "phase" in config:
            pag = (
                "method" in config["phase"]
                and config["phase"]["method"] == _NabuPhaseMethod.PAGANIN.value
            )
            ctf = (
                "method" in config["phase"]
                and config["phase"]["method"] == _NabuPhaseMethod.CTF.value
            )
            if pag or ctf:
                if "delta_beta" in config["phase"]:
                    if not numpy.isscalar(config["phase"]["delta_beta"]):
                        if len(config["phase"]["delta_beta"]) > 1:
                            raise ValueError(
                                "expects at most one value for 'delta_beta'"
                            )
                        else:
                            config["phase"]["delta_beta"] = config["phase"][
                                "delta_beta"
                            ][0]
                    db = round(float(config["phase"]["delta_beta"]))
        # retrieve axis used
        axis = config.get("reconstruction", {}).get("slice_plane", "XY")

        # handle output
        if "output" in config:
            _file_name = self._get_file_basename_reconstruction(
                pag=pag, db=db, ctf=ctf, axis=axis
            )
            config["output"]["file_prefix"] = _file_name
            location, location_cfg_files = get_output_folder_from_scan(
                mode=ProcessDataOutputDirMode.from_value(
                    config["output"].get(
                        "output_dir_mode", ProcessDataOutputDirMode.OTHER
                    )
                ),
                nabu_location=config["output"].get("location", None),
                scan=self.scan,
                file_basename=_file_name,
                file_format=config["output"].get(
                    "file_format", NabuOutputFileFormat.HDF5
                ),
            )
            # add reconstruction path to the list. scan `reconstruction_paths` register all the existing path where
            # reconstruction are saved in order to be able to browse them all
            self.scan.add_reconstruction_path(location)
            config["output"]["location"] = location
        else:
            # don't think this could ever happen
            location_cfg_files = self.scan.path
        # handle preproc
        if "preproc" not in config:
            config["preproc"] = {}
        if self.scan.intensity_normalization.method is INormMethod.NONE:
            config["preproc"]["sino_normalization"] = ""
        else:
            config["preproc"][
                "sino_normalization"
            ] = self.scan.intensity_normalization.method.value

        extra_infos = self.scan.intensity_normalization.get_extra_infos()

        nabu_cfg_folder = os.path.join(
            location_cfg_files, settings.NABU_CFG_FILE_FOLDER
        )
        os.makedirs(nabu_cfg_folder, exist_ok=True)

        # configuration file and nabu_tomwer_serving_hatch must be in the same folder
        serving_hatch_file = os.path.join(
            nabu_cfg_folder, settings.NABU_TOMWER_SERVING_HATCH
        )

        source = extra_infos.get("source", INormSource.NONE)
        source = INormSource.from_value(source)

        if source is INormSource.NONE:
            pass
        elif source is INormSource.MANUAL_SCALAR:
            if "value" not in extra_infos:
                raise KeyError(
                    "value should be provided in extra)infos for scalar defined manually"
                )
            else:
                # check if the dataset has already been saved once and if we can reuse it
                dataset_url = extra_infos.get("dataset_created_by_tomwer", None)
                if dataset_url is not None:
                    # if an url exists insure we can access it
                    dataset_url = DataUrl(path=dataset_url)
                    if os.path.exists(dataset_url.file_path()):
                        with open_hdf5(dataset_url.file_path()) as h5f:
                            if dataset_url.data_path() not in h5f:
                                dataset_url = None
                    else:
                        dataset_url = None
                # if unable toi reuse an existing url them dump the value
                if dataset_url is None:
                    value = extra_infos["value"]
                    if isinstance(value, (tuple, list)):
                        value = numpy.asarray(value)
                    dataset_url = dump_normalization_array_for_nabu(
                        scan=self.scan,
                        array=value,
                        output_file=serving_hatch_file,
                    )
                    extra_infos.update(
                        {"dataset_created_by_tomwer": dataset_url.path()}
                    )
                    self.scan.intensity_normalization.set_extra_infos(extra_infos)

                config["preproc"]["sino_normalization_file"] = dataset_url.path()
        elif source is INormSource.DATASET:
            url = extra_infos["dataset_url"]
            if isinstance(url, DataUrl):
                config["preproc"]["sino_normalization_file"] = url.path()
            elif isinstance(url, str):
                config["preproc"]["sino_normalization_file"] = url
            else:
                raise TypeError(
                    f"dataset_url is expected to be an instance of DataUrl or str representing a DataUrl. Not {type(url)}"
                )
        else:
            raise NotImplementedError(f"source type {source.value} is not handled")
        return config, nabu_cfg_folder

    def _get_file_basename_reconstruction(self, pag, db, ctf, axis):
        """return created file base name"""
        raise NotImplementedError("Base class")

    def cancel(self):
        self._cancelled = True
        if self._process:
            # kill childs processes
            try:
                parent = psutil.Process(self._process.pid)
            except psutil.NoSuchProcess:
                pass
            else:
                # TODO: see with Pierre. But from my point of view this
                # might be handled by nabu it self....
                childrens = parent.children(recursive=True)
                for child in childrens:
                    try:
                        child.send_signal(signal.SIGKILL)
                    except psutil.NoSuchProcess:
                        pass
                    else:
                        child.wait()
                self._process.kill()
                self._process.wait()
            finally:
                self._process = None


def dump_normalization_array_for_nabu(
    scan: TomwerScanBase, output_file: str, array: Union[numpy.ndarray, float, int]
) -> DataUrl:
    if not isinstance(array, (numpy.ndarray, float, int)):
        raise TypeError(
            f"array is expected to be a numpy array or a scalar and not {type(array)}"
        )
    # save the value to a dedicated path in "nabu_tomwer_serving_hatch"
    if isinstance(scan, NXtomoScan):
        entry_path = scan.entry
    elif isinstance(scan, EDFTomoScan):
        entry_path = "entry"
    else:
        raise TypeError
    with HDF5File(output_file, mode="a") as h5f:
        serving_hatch_data_path = None
        # create a unique dataset path to avoid possible conflicts
        while serving_hatch_data_path is None or serving_hatch_data_path in h5f:
            serving_hatch_data_path = "/".join([entry_path, str(uuid.uuid1())])
        # adapt value to what nabues expects.
        if isinstance(array, (float, int)) or (
            isinstance(array, numpy.ndarray) and array.ndim == 1 and len(array) == 1
        ):
            dim_1 = scan.dim_1
            array = numpy.asarray(
                numpy.asarray([array] * len(scan.projections) * dim_1)
            )
            array = array.reshape(len(scan.projections), dim_1)
        elif isinstance(array, numpy.ndarray) and array.ndim == 1:
            dim_1 = scan.dim_1
            array = numpy.repeat(array, dim_1).reshape(len(array), dim_1)

        h5f[serving_hatch_data_path] = array
    file_path = os.path.join(
        settings.NABU_CFG_FILE_FOLDER, settings.NABU_TOMWER_SERVING_HATCH
    )
    return DataUrl(
        file_path=file_path,
        data_path=serving_hatch_data_path,
        scheme="silx",
    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"


import os
import shutil
import tempfile

import h5py
import numpy.random
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.gui.reconstruction.axis import AxisWindow
from tomwer.gui.reconstruction.axis.radioaxis import _ManualFramesSelection
from tomwer.synctools.axis import QAxisRP
from tomwer.tests.datasets import TomwerCIDatasets
from tomwer.tests.utils import skip_gui_test
from tomwer.core.utils.scanutils import MockNXtomo


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestWindowAxis(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.recons_params = QAxisRP()
        self._window = AxisWindow(axis_params=self.recons_params)
        self.scan = MockNXtomo(
            scan_path=os.path.join(self.tempdir, "nx_tomo.nx"),
            n_proj=10,
            n_ini_proj=10,
            scan_range=180,
            dim=20,
            energy=12.3,
        ).scan
        self._window.show()
        self.qWaitForWindowExposed(self._window)

    def tearDown(self):
        shutil.rmtree(self.tempdir)
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()

    def testSetImagesNumpyArray(self):
        """Test the setImages function"""
        radio_axis = self._window._axisWidget._radioAxis
        radio_axis_plot = self._window._axisWidget._radioAxis.getPlot()
        numpy.testing.assert_array_equal(
            radio_axis_plot.getPlot().getActiveImage().getData(), numpy.empty((0, 0, 4))
        )
        imgA = numpy.random.random((100, 100))
        imgB = numpy.random.random((100, 100))
        radio_axis.setImages(imgA=imgA, imgB=imgB, flipB=True)
        self.qapp.processEvents()
        # make sur the image has been set
        assert radio_axis_plot.getPlot().getImage(legend="image1").getData().shape == (
            100,
            100,
        )

    def testSetScan(self):
        """Test the setScan function"""
        radio_axis_plot = self._window._axisWidget._radioAxis.getPlot()
        numpy.testing.assert_array_equal(
            radio_axis_plot.getPlot().getActiveImage().getData(), numpy.empty((0, 0, 4))
        )
        self._window.setScan(self.scan)
        self.qapp.processEvents()
        # make sur the image has been set
        assert radio_axis_plot.getPlot().getImage(legend="image1").getData().shape == (
            20,
            20,
        )

    def testShiftButtons(self):
        """Test that the 'left', 'right', ... buttons and the shift steps are
        correctly working"""
        imgA = numpy.random.random((100, 100))
        imgB = numpy.random.random((100, 100))
        radio_axis = self._window._axisWidget._radioAxis
        radio_axis.setImages(imgA=imgA, imgB=imgB, flipB=False)
        self.qapp.processEvents()
        self.assertTrue(radio_axis.getXShift() == 0)
        self.assertTrue(radio_axis.getYShift() == 0)
        self.assertTrue(
            radio_axis.getShiftStep() == 1.0,
            f"shift is {radio_axis.getShiftStep()} when 1.0 expected",
        )
        radio_axis._controlWidget._shiftControl._leftButton.click()
        self.qapp.processEvents()
        self.assertTrue(radio_axis.getXShift() == -1)
        self.assertTrue(radio_axis.getYShift() == 0)
        radio_axis._controlWidget._shiftControl._rightButton.click()
        self.qapp.processEvents()
        self.assertTrue(radio_axis.getXShift() == 0)
        self.assertTrue(radio_axis.getYShift() == 0)
        radio_axis._controlWidget._displacementSelector._fineButton.click()
        self.qapp.processEvents()
        self.assertTrue(radio_axis.getShiftStep() == 0.1)
        radio_axis._controlWidget._shiftControl._topButton.click()
        self.qapp.processEvents()
        self.assertTrue(radio_axis.getXShift() == 0)
        self.assertTrue(radio_axis.getYShift() == 0.1)
        radio_axis.setShiftStep(0.2)
        self.assertTrue(radio_axis.getShiftStep() == 0.2)
        radio_axis._controlWidget._shiftControl._bottomButton.click()
        self.qapp.processEvents()
        self.assertTrue(radio_axis.getXShift() == 0)
        self.assertTrue(radio_axis.getYShift() == -0.1)

    def testAxisObjectGlobal(self):
        """Test that the GUI change according to the Axis object"""
        axis_obj = QAxisRP()
        self._window.setReconsParams(axis_obj)
        self.qapp.processEvents()
        self.assertEqual(self._window._axisWidget._radioAxis.getMode(), axis_obj.mode)
        # TODO: change the mode to see if the object is modify

    def testAxisObjectManual(self):
        """Test that the GUI change according to the Axis object"""
        axis_params = QAxisRP()
        axis_params.mode = AxisMode.manual
        axis_params.set_value_ref_tomwer(-6.0)
        # note : for now only the xshift is managed !
        self._window.setReconsParams(axis_params)
        radio_axis = self._window._axisWidget._radioAxis
        self.assertEqual(radio_axis.getXShift(), -6.0)
        self.assertEqual(radio_axis.getYShift(), 0.0)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestWindowsAxisSetScan(TestCaseQt):
    """Test that updates are correctly done in the case scan has some
    knowledge on"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self.recons_params = QAxisRP()
        self.widget = AxisWindow(axis_params=self.recons_params)
        self.scan_dir = tempfile.mkdtemp()
        # create dataset
        self.master_file = os.path.join(self.scan_dir, "frm_edftomomill_twoentries.nx")
        shutil.copyfile(
            TomwerCIDatasets.get_dataset(
                "h5_datasets/frm_edftomomill_twoentries.nx",
            ),
            self.master_file,
        )
        self.scan = NXtomoScan(scan=self.master_file, entry="entry0000")

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None

    def patch_fov(self, value: str):
        with h5py.File(self.scan.master_file, mode="a") as h5s:
            for entry in ("entry0000", "entry0001"):
                entry_node = h5s[entry]
                if "instrument/detector/field_of_view" in entry_node:
                    del entry_node["instrument/detector/field_of_view"]
                entry_node["instrument/detector/field_of_view"] = value


def test__getClosestAssociatedAngle():
    """
    test _getClosestAssociatedAngle function
    """
    assert (
        _ManualFramesSelection._getClosestAssociatedAngle(angle=12.3, angles=(12.3,))
        == 12.3
    )
    assert (
        _ManualFramesSelection._getClosestAssociatedAngle(
            angle=12.3, angles=(2.06, 10.3, 100.6)
        )
        == 10.3
    )
    assert (
        _ManualFramesSelection._getClosestAssociatedAngle(
            angle=100.2, angles=(2.06, 10.3, 100.6)
        )
        == 100.6
    )
    assert (
        _ManualFramesSelection._getClosestAssociatedAngle(angle=12.3, angles=()) is None
    )
    assert (
        _ManualFramesSelection._getClosestAssociatedAngle(
            angle=300.2, angles=(2.06, 10.3, 100.6)
        )
        == 100.6
    )
    assert (
        _ManualFramesSelection._getClosestAssociatedAngle(
            angle=-12, angles=(2.06, 10.3, 100.6)
        )
        == 2.06
    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
material for radio and sinogram normalization
"""


__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "25/06/2021"


import typing

from silx.utils.enum import Enum as _Enum
from tomoscan.normalization import Method


class _ValueSource(_Enum):
    MONITOR = "intensity monitor"
    MANUAL_ROI = "manual ROI"
    AUTO_ROI = "automatic ROI"
    DATASET = "from dataset"
    MANUAL_SCALAR = "scalar"
    NONE = "none"


class _ValueCalculationFct(_Enum):
    MEAN = "mean"
    MEDIAN = "median"


class _DatasetScope(_Enum):
    LOCAL = "local"
    GLOBAL = "global"


class _DatasetInfos:
    def __init__(self):
        self._scope = _DatasetScope.GLOBAL
        self._file_path = None
        self._data_path = None

    @property
    def scope(self) -> _DatasetScope:
        return self._scope

    @scope.setter
    def scope(self, scope: typing.Union[str, _DatasetScope]):
        self._scope = _DatasetScope.from_value(scope)

    @property
    def file_path(self):
        return self._file_path

    @file_path.setter
    def file_path(self, file_path):
        self._file_path = file_path

    @property
    def data_path(self):
        return self._data_path

    @data_path.setter
    def data_path(self, data_path: str):
        self._data_path = data_path


class _ROIInfo:
    def __init__(self, x_min=None, x_max=None, y_min=None, y_max=None):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max


class SinoNormalizationParams:
    """Information regarding the intensity normalization to be done"""

    def __init__(self, method=Method.NONE, source=_ValueSource.NONE, extra_infos=None):
        self._method = Method.NONE
        self._source = _ValueSource.NONE
        self._extra_infos = {}

        self.method = method
        self.extra_infos = extra_infos if extra_infos is not None else {}
        self.source = source

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method: typing.Union[str, Method, None]):
        if method is None:
            method = Method.NONE
        self._method = Method.from_value(method)

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        if source is None:
            source = _ValueSource.NONE
        self._source = _ValueSource.from_value(source)

    @property
    def extra_infos(self):
        return self._extra_infos

    @extra_infos.setter
    def extra_infos(self, extra_infos: dict):
        if not isinstance(extra_infos, dict):
            raise TypeError("extra infos is expected to be a dictionary")
        elif "method" in extra_infos:
            raise KeyError("'method' is a reserved key name")
        else:
            self._extra_infos = extra_infos

    def to_dict(self):
        _dict = self._extra_infos
        _dict["method"] = self.method.value
        _dict["source"] = self.source.value
        return _dict

    @staticmethod
    def from_dict(dict_):
        params = SinoNormalizationParams()
        params.load_from_dict(dict_=dict_)
        return params

    def load_from_dict(self, dict_):
        tmp_dict = dict_.copy()
        if "method" in tmp_dict:
            self.method = tmp_dict["method"]
            del tmp_dict["method"]
        if "source" in tmp_dict:
            self.source = tmp_dict["source"]
            del tmp_dict["source"]
        self.extra_infos = tmp_dict



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"


import functools
from typing import Union

from silx.gui import qt
from silx.utils.enum import Enum as _Enum

from tomwer.gui import icons
from tomwer.utils import docstring


class FlowDirection(_Enum):
    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"


class _ProcessDescWidget(qt.QWidget):
    """Widget to describe the process"""

    ICON_WIDTH = 16

    def __init__(self, parent, process, icon, direction, locked, show_lock_state=True):
        assert isinstance(direction, FlowDirection)
        super(_ProcessDescWidget, self).__init__(parent=parent)
        self.__active = False

        if direction is FlowDirection.VERTICAL:
            self.setLayout(qt.QHBoxLayout())
        else:
            self.setLayout(qt.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        # lock icon
        self._lockIcon = qt.QLabel("", parent=self)
        self._lockIcon.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self._lockIcon.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignTop)

        if locked:
            locked_icon = icons.getQIcon("locked")
        else:
            locked_icon = icons.getQIcon("unlocked")

        self._lockIcon.setPixmap(locked_icon.pixmap(20, 20))
        self._lockIcon.setVisible(show_lock_state)

        # process name + process icon
        self.main_widget = qt.QWidget(parent=self)
        self.main_widget.setLayout(qt.QHBoxLayout())

        self._label = qt.QLabel(str(process), self)
        self._label.setAlignment(qt.Qt.AlignCenter)
        self._label.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.main_widget.layout().addWidget(self._label)
        if icon is not None:
            self._frame = qt.QLabel(self)
            self._frame.setPixmap(
                icon.pixmap(
                    qt.QSize(
                        _ProcessDescWidget.ICON_WIDTH, _ProcessDescWidget.ICON_WIDTH
                    )
                )
            )
            self.main_widget.layout().addWidget(self._frame)

        # main_widget.layout().addWidget(icon_label)
        if direction is FlowDirection.VERTICAL:
            self.layout().addWidget(self.main_widget)
            self.layout().addWidget(self._lockIcon)
        else:
            self.layout().addWidget(self._lockIcon)
            self.layout().addWidget(self.main_widget)
        # expose API
        self.process = process

    def setActive(self, active: bool) -> None:
        """
        set the widget as active or inactive. This will put or remove the focus
        on.

        :param active:
        """
        self.__active = active
        if active:
            stylesheet = "border: 1px solid red"
        else:
            stylesheet = ""
        self.main_widget.setStyleSheet(stylesheet)
        if hasattr(self, "_frame"):
            self._frame.setStyleSheet("border: 0px")
        self._label.setStyleSheet("border: 0px")

    def isActive(self) -> bool:
        """

        :return: is the widget active / focus / under edition
        """
        return self.__active


class _ProcessDescContainer(qt.QWidget):
    """simple docker for the process description

    :param parent:
    :param process: any object implementing the `__str__` interface
    :param icon:
    :raises: TypeError if process does not implements __str__
    """

    sigWidgetActivated = qt.Signal()
    """Signal emitted when the contained widget become the active process."""
    sigWidgetDeactivated = qt.Signal()
    """Signal emitted when the contained widget is deactivated (not the
    active process anymore)"""

    def __init__(
        self, parent, process, icon, direction, draggable, show_lock_state=True
    ):
        if not hasattr(process, "__str__"):
            raise TypeError("name should implement `__str__`")
        super(_ProcessDescContainer, self).__init__(parent=parent)
        self._widget = _ProcessDescWidget(
            parent=self,
            process=process,
            icon=icon,
            direction=direction,
            locked=not draggable,
            show_lock_state=show_lock_state,
        )
        self.setLayout(qt.QVBoxLayout())
        self.setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._widget)
        # expose API
        self.process = self._widget.process
        self._draggable = draggable

    def event(self, QEvent):
        if (
            QEvent.type() == qt.QEvent.MouseButtonPress
            and QEvent.button() == qt.Qt.LeftButton
        ):
            self._setIsOnFocus()
        return super(_ProcessDescContainer, self).event(QEvent)

    def _setIsOnFocus(self):
        if self._widget.isActive():
            self.sigWidgetDeactivated.emit()
        else:
            self.sigWidgetActivated.emit()

    @docstring(_ProcessDescWidget)
    def setActive(self, active):
        self._widget.setActive(active=active)
        if active:
            self.sigWidgetActivated.emit()
        else:
            self.sigWidgetDeactivated.emit()

    @docstring(_ProcessDescWidget)
    def isActive(self):
        return self._widget.isActive()


class FlowCanvas(qt.QWidget):
    """
    Widget to describe a flow of process - pipeline

    :param direction: direction can be vertical or horizontal
    :type direction: Union[str,FlowDirection]
    """

    sigWidgetActivated = qt.Signal()
    """Signal emitted when the contained widget become the active process."""
    sigWidgetDeactivated = qt.Signal()
    """Signal emitted when the contained widget is deactivated (not the
    active process anymore)"""

    def __init__(
        self, parent, direction: Union[str, FlowDirection], show_lock_state=True
    ) -> None:
        qt.QWidget.__init__(self, parent=parent)
        self._direction = FlowDirection.from_value(direction)
        if self._direction is FlowDirection.HORIZONTAL:
            layout = qt.QHBoxLayout()
        elif self._direction is FlowDirection.VERTICAL:
            layout = qt.QVBoxLayout()
        else:
            raise ValueError("direction not managed")
        self.setLayout(layout)
        self.setWindowFlags(qt.Qt.Widget)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self._dockers = {}
        self._process_id = 0
        self._show_lock_state = show_lock_state

        if self._direction is FlowDirection.HORIZONTAL:
            self._dockArea = qt.Qt.TopDockWidgetArea
        elif self._direction is FlowDirection.VERTICAL:
            self._dockArea = qt.Qt.LeftDockWidgetArea
        else:
            raise ValueError("direction not managed")

    def getFlow(self) -> list:
        """Return the list of process following the treatment flow"""
        widgetXPos = {}
        # retrieve each docker position
        for process_id, dockWidget in self._dockers.items():
            if self._direction is FlowDirection.HORIZONTAL:
                widgetXPos[process_id] = dockWidget.pos().x()
            else:
                widgetXPos[process_id] = dockWidget.pos().y()
        # order the widget / process according to
        res = []
        for widgetIndex, _ in sorted(widgetXPos.items(), key=lambda item: item[1]):
            res.append(self._dockers[widgetIndex].process)
        return res

    def addProcess(self, process, icon=None, draggable=True) -> int:
        """
        Add a process to the flow with the name 'process' and an optional icon

        :param process_name: any object. Should implement the __str__ interface
                             to define the name if no icon is given
        :return: process id in the process flow
        :rtype: int
        """
        if icon is None:
            assert hasattr(process, "__str__")
        widget = _ProcessDescContainer(
            parent=self,
            process=process,
            icon=icon,
            direction=self._direction,
            draggable=draggable,
            show_lock_state=self._show_lock_state,
        )

        self.layout().addWidget(widget)
        process_id = self._process_id
        self._dockers[self._process_id] = widget
        self._process_id += 1

        # connect signals / slots
        widget.sigWidgetActivated.connect(
            functools.partial(self._activeProcessChanged, process_id)
        )
        widget.sigWidgetDeactivated.connect(
            functools.partial(self._activeProcessChanged, None)
        )

        return process_id

    def _activeProcessChanged(self, active_process):
        for process_id, dock in self._dockers.items():
            old = dock.blockSignals(True)
            dock.setActive(active_process == process_id)
            dock.blockSignals(old)
        if active_process is None:
            self.sigWidgetDeactivated.emit()
        else:
            self.sigWidgetActivated.emit()

    def removeProcess(self, process_id: int) -> None:
        """

        :param int process_id: id of the process to remove from the flow
        """
        if process_id in self._dockers:
            dockWidget = self._dockers[process_id]
            self.layout().removeWidget(dockWidget)
            del self._dockers[process_id]

    def clear(self):
        processes_id = self._dockers.keys()
        for process_id in processes_id:
            self.removeProcess(process_id=process_id)

    def setProcesses(self, processes, icons, draggability) -> None:
        """

        :param Iterable processes: processes of the flow
        :param Iterable icons: icons associated to the flow
        :param Iterable draggability: is the process draggable or not
        :return:
        """
        for process, icon, draggable in zip(processes, icons, draggability):
            self.addProcess(process=process, draggable=draggable, icon=icon)

    def clearActiveProcess(self) -> None:
        """
        Set all process to unactive
        """
        for dock in self._dockers.values():
            old = dock.blockSignals(True)
            dock.setActive(False)
            dock.blockSignals(old)

    def hasProcessFocus(self) -> bool:
        """

        :return: True if the canvas has a process active / focus ...
        :rtype: bool
        """
        return self.getProcessFocused() is not None

    def getProcessFocused(self) -> Union[None, int]:
        """

        :return: id of the process focus or None if no process focus
        """
        for process_id, dock in self._dockers.items():
            if dock.isActive():
                return process_id
        return None

    def _removeActiveProcess(self):
        process_id = self.getProcessFocused()
        if process_id is not None:
            self.removeProcess(process_id=process_id)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "20/04/2020"


import os
from flufl.lock import Lock

from tomwer.core.utils.Singleton import singleton

_FILE_LOCKERS = {}


def get_lock_file_path(process_file_path):
    process_file_path = os.path.abspath(process_file_path)
    parts = process_file_path.split(os.sep)
    if len(parts) == 0:
        # file without any name ???!!!
        raise RuntimeError
    elif len(parts) == 1:
        file_path = ""
    else:
        file_path = os.path.join(*parts[:-1])
    lock_name = "." + parts[-1] + ".flufllock"
    return os.path.join(file_path, lock_name)


@singleton
class FileLockerManager:
    """Insure that for each file we will provide at most one locker"""

    def __init__(self):
        self.__lockers = {}

    def clear_locker(self, file_):
        if file_ in self.__lockers:
            del self.__lockers[file_]

    @staticmethod
    def get_lock(file_name):
        def get_lock_file_path(file_path):
            file_path = os.path.abspath(file_path)
            parts = file_path.split(os.sep)
            if len(parts) == 0:
                # file without any name ???!!!
                raise RuntimeError
            elif len(parts) == 1:
                file_path = ""
            else:
                file_path = os.sep.join(parts[:-1])
            lock_name = "." + parts[-1] + ".flufllock"
            return os.path.join(file_path, lock_name)

        lock_file_path = get_lock_file_path(file_name)
        # if not os.path.exists(lock_file_path):
        #     from pathlib import Path
        #     Path(lock_file_path).touch()
        if lock_file_path not in _FILE_LOCKERS:
            _FILE_LOCKERS[lock_file_path] = Lock(lock_file_path, default_timeout=3)
        return _FILE_LOCKERS[lock_file_path]



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/05/2018"

import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output, OWBaseWidget
from silx.gui import qt

import tomwer.core.process.control.scanselector
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.scanselectorwidget import ScanSelectorWidget

logger = logging.getLogger(__name__)


class DataSelectorOW(OWBaseWidget, openclass=True):
    name = "scan selector"
    id = "orange.widgets.tomwer.scanselector"
    description = (
        "List all received scan. Then user can select a specific"
        "scan to be passed to the next widget."
    )
    icon = "icons/scanselector.svg"
    priority = 42
    keywords = ["tomography", "selection", "tomwer", "scan", "data"]

    ewokstaskclass = tomwer.core.process.control.scanselector._ScanSelectorPlaceHolder

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _scanIDs = Setting(list())

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, multiple=True)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """ """
        super().__init__(parent)

        self.widget = ScanSelectorWidget(parent=self)
        self.widget.setWindowFlags(qt.Qt.Widget)
        self._loadSettings()

        self.widget.sigUpdated.connect(self._updateSettings)
        self.widget.sigSelectionChanged.connect(self.changeSelection)
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)

    @Inputs.data
    def addScan(self, scan, *args, **kwargs):
        self.add(scan)

    def add(self, scan):
        if scan is not None:
            self.widget.add(scan=scan)

    def changeSelection(self, list_objs):
        if list_objs is None:
            return
        for obj_id in list_objs:
            item = self.widget.dataList._myitems.get(obj_id, None)
            if item:
                tomo_obj = item.data(qt.Qt.UserRole)
                assert isinstance(tomo_obj, TomwerScanBase)
                self.Outputs.data.send(tomo_obj)
            else:
                logger.error("%s not found in scan ids" % obj_id)

    def send(self):
        """send output signals for each selected items"""
        sItem = self.widget.dataList.selectedItems()
        if sItem and len(sItem) >= 1:
            selection = [
                _item.data(qt.Qt.UserRole).get_identifier().to_str() for _item in sItem
            ]
            self.changeSelection(list_objs=selection)

    def _loadSettings(self):
        for scan in self._scanIDs:
            assert isinstance(scan, str)
            # kept for backward compatibility since 0.11. To be removed on the future version.
            if "@" in scan:
                entry, file_path = scan.split("@")
                nxtomo_scan = NXtomoScan(entry=entry, scan=file_path)
                self.addScan(nxtomo_scan)
            else:
                self.addScan(scan)

    def _updateSettings(self):
        self._scanIDs = []
        for scan in self.widget.dataList._myitems:
            self._scanIDs.append(scan)

    def keyPressEvent(self, event):
        """
        To shortcut orange and make sure the `delete` key will be interpreted we need to overwrite this function
        """
        if event.key() == qt.Qt.Key_Delete:
            self.widget.removeSelectedDatasets()
        else:
            super().keyPressEvent(event)

    # expose API
    def setActiveScan(self, data):
        self.widget.setActiveData(data=data)

    def selectAll(self):
        return self.widget.selectAll()

    def n_scan(self):
        return self.widget.n_scan()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import signal
import sys
import functools

import silx
from silx.gui import qt

from nabu.utils import Progress
from nabu.stitching.config import StitchingType

from tomwer.core.settings import SlurmSettingsMode
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.core.process.stitching.nabustitcher import StitcherTask
from tomwer.io.utils.tomoobj import get_tomo_objs_instances
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.stitching.stitching import ZStitchingWindow as _ZStitchingWindow
from tomwer.gui.stitching import action as stitching_action
from tomwer.gui.cluster.slurm import SlurmSettingsWidget
from tomwer.core.scan.scanbase import TomwerScanBase

logging.basicConfig(level=logging.WARNING)

_logger = logging.getLogger(__name__)


class MainWidget(qt.QTabWidget):
    DEFAULT_NB_WORKERS = 10

    def __init__(
        self,
        parent=None,
    ) -> None:
        super().__init__(parent)
        self._stitchingConfigWindow = _ZStitchingWindow(
            parent=self, with_configuration_action=False
        )
        self.addTab(self._stitchingConfigWindow, "stitching config")
        settingsClass = SlurmSettingsMode.get_settings_class(
            SlurmSettingsMode.STITCHING
        )
        self._slurmConfig = SlurmSettingsWidget(
            parent=self, n_gpu=settingsClass.N_GPUS_PER_WORKER, jobLimitation=None
        )
        self._slurmConfig.setConfiguration(
            {
                "cpu-per-task": settingsClass.N_CORES_PER_TASK,
                "n_tasks": settingsClass.N_TASKS,
                "n_jobs": settingsClass.N_JOBS,
                "memory": settingsClass.MEMORY_PER_WORKER,
                "partition": settingsClass.PARTITION,
                "n_gpus": settingsClass.N_GPUS_PER_WORKER,
                "job_name": settingsClass.PROJECT_NAME,
                "python_venv": settingsClass.PYTHON_VENV,
            }
        )

        self._stitchingConfigWindow.setCallbackToGetSlurmConfig(
            self._slurmConfig.getConfiguration
        )
        self._stitchingConfigWindow.setCallbackToSetSlurmConfig(
            self._slurmConfig.setConfiguration
        )
        self.addTab(self._slurmConfig, "slurm config")
        # add an option to clearly activate / deactivate slurm config
        slurm_config_idx = self.indexOf(self._slurmConfig)
        self._slurmCB = qt.QCheckBox(self)
        self.tabBar().setTabButton(
            slurm_config_idx,
            qt.QTabBar.LeftSide,
            self._slurmCB,
        )
        self._slurmCB.setChecked(True)

        # set up
        self._slurmConfig.setNWorkers(self.DEFAULT_NB_WORKERS)

    def close(self):
        self._stitchingConfigWindow.close()
        # requested for the waiting plot update
        super().close()

    def keyPressEvent(self, event):
        """
        To shortcut orange and make sure the `F5` <=> refresh stitching preview
        """
        modifiers = event.modifiers()
        key = event.key()

        if key == qt.Qt.Key_F5:
            self._stitchingConfigWindow._trigger_update_preview()
        elif key == qt.Qt.Key_O and modifiers == qt.Qt.KeyboardModifier.ControlModifier:
            self._stitchingConfigWindow._loadSettings()
        elif key == qt.Qt.Key_S and modifiers == qt.Qt.KeyboardModifier.ControlModifier:
            self._stitchingConfigWindow._saveSettings()

    def getStitchingConfiguration(self) -> dict:
        return self._stitchingConfigWindow.getConfiguration()

    def getClusterConfig(self) -> dict:
        return self._slurmConfig.getConfiguration()

    # expose API
    def addTomoObj(self, *args, **kwargs):
        self._stitchingConfigWindow.addTomoObj(*args, **kwargs)

    def setStitchingType(self, stitching_type):
        self._stitchingConfigWindow.setStitchingType(stitching_type)

    def loadSettings(self, config_file):
        self._stitchingConfigWindow._loadSettings(config_file)


class _StitcherThread(qt.QThread):
    def __init__(self, parent, stitching_config, cluster_config) -> None:
        super().__init__(parent)
        assert isinstance(stitching_config, dict)
        assert isinstance(cluster_config, dict)
        self._stitching_config = stitching_config
        self._cluster_config = cluster_config

    def run(self):
        task = StitcherTask(
            inputs={
                "stitching_config": self._stitching_config,
                "cluster_config": self._cluster_config,
                "progress": Progress("stitching"),
                "serialize_output_data": False,
            },
        )

        if self._cluster_config in (None, {}):
            self._runLocally(task)
        else:
            self._runRemotly(task)

    def _runLocally(self, task):
        task.run()

    def _runRemotly(self, task):
        task.run()
        # wait for processing to be finished
        for future in task.outputs.future_tomo_obj.futures:
            future.result()
        # note: there is also a callback to aggregate result once all are finished.
        # so at this stage the final "nxtomo" or "volume" might not exists yet
        print("Add slurm job finished.")


class MainWindow(qt.QDialog):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        # menu
        self._menuBar = qt.QMenuBar(self)
        self.layout().addWidget(self._menuBar)
        self._menu = qt.QMenu("&Stitching")
        self._menuBar.addMenu(self._menu)
        # save / load configuration
        self._loadConfigurationAction = stitching_action.LoadConfigurationAction(
            self, "&Load configuration"
        )
        self._menu.addAction(self._loadConfigurationAction)
        self._saveConfigurationAction = stitching_action.SaveConfigurationAction(
            self, "&Save configuration"
        )
        self._menu.addAction(self._saveConfigurationAction)
        # separator
        self._menu.addSeparator()
        # add tomo object
        self._addTomoObjectAction = stitching_action.AddTomoObjectAction(
            self, "&Add tomo object (volume or NXtomo)"
        )
        self._menu.addAction(self._addTomoObjectAction)

        # main window
        self.setWindowTitle("tomwer z-stitching")
        self._mainWindow = MainWidget(parent=self)
        self.layout().addWidget(self._mainWindow)

        # button
        types = qt.QDialogButtonBox.Apply
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)
        self._buttons.button(qt.QDialogButtonBox.Apply).setText("Launch stitching")

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Apply).clicked.connect(self.accept)
        self._loadConfigurationAction.triggered.connect(
            functools.partial(
                self._mainWindow._stitchingConfigWindow._loadSettings, file_path=None
            )
        )
        self._saveConfigurationAction.triggered.connect(
            functools.partial(
                self._mainWindow._stitchingConfigWindow._saveSettings, file_path=None
            )
        )
        self._addTomoObjectAction.triggered.connect(self._callbackAddTomoObj)

    def _callbackAddTomoObj(self, *args, **kwargs):
        """move interface to the z-ordered list and call the 'add tomo obj' callback"""
        orderedListWidget = (
            self._mainWindow._stitchingConfigWindow._widget._mainWidget._zOrderedList
        )
        self._mainWindow._stitchingConfigWindow._widget._mainWidget.setCurrentWidget(
            orderedListWidget
        )
        orderedListWidget._callbackAddTomoObj()

    def close(self):
        self._mainWindow.close()
        # requested for the waiting plot update
        super().close()

    def reject(self):
        self.close()
        super().reject()

    def accept(self):
        self.process_stitching()

    def process_stitching(self):
        self._processingThread = _StitcherThread(
            parent=self,
            stitching_config=self.getStitchingConfiguration(),
            cluster_config=self.getClusterConfig(),
        )
        self._processingThread.start()

    def disableSlurmconfig(self):
        self._mainWindow._slurmCB.setChecked(False)
        self._mainWindow._slurmCB.setCheckable(False)

        slurm_config_idx = self._mainWindow.indexOf(self._mainWindow._slurmConfig)

        self._mainWindow.setTabEnabled(
            slurm_config_idx,
            False,
        )
        self._mainWindow.setTabToolTip(
            slurm_config_idx,
            "not available because unable to find slurm command (computer is probably not a slurm client ?",
        )

    # expose API
    def addTomoObj(self, *args, **kwargs):
        self._mainWindow.addTomoObj(*args, **kwargs)

    def getClusterConfig(self) -> dict:
        if self._mainWindow._slurmCB.isChecked():
            return self._mainWindow.getClusterConfig()
        else:
            return {}

    def getStitchingConfiguration(self) -> dict:
        return self._mainWindow.getStitchingConfiguration()

    def setStitchingType(self, stitching_type):
        self._mainWindow.setStitchingType(stitching_type)

    def loadSettings(self, config_file: str):
        self._mainWindow.loadSettings(config_file)


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "tomo_objs",
        help="All tomo objects to be stitched together",
        nargs="*",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    parser.add_argument(
        "--config",
        "--config-file",
        dest="config_file",
        default=None,
        help="Provide stitching configuration file to load parameters from it",
    )
    parser.add_argument(
        "--use-opengl-plot",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--check-scans",
        help="If True will check scans before adding them. This check scan validity (virtual dataset contains data...)",
        action="store_true",
        default=False,
    )
    options = parser.parse_args(argv[1:])

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])
    splash = getMainSplashScreen()
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    window = MainWindow()
    window.setWindowFlags(qt.Qt.Window)
    window.setWindowTitle("tomwer: z-stitching")
    window.setWindowIcon(icons.getQIcon("tomwer"))

    window.show()
    from sluurp.utils import has_sbatch_available

    # load configuration is some provided
    if options.config_file is not None:
        window.loadSettings(options.config_file)

    if not has_sbatch_available():
        window.disableSlurmconfig()

    tomo_objs, (has_scans, has_vols) = get_tomo_objs_instances(options.tomo_objs)
    for tomo_obj in tomo_objs:
        if isinstance(tomo_obj, TomwerScanBase):
            tomo_obj.set_check_behavior(
                run_check=options.check_scans
            )  # speed up processing
        window.addTomoObj(tomo_obj=tomo_obj)
    if has_scans and has_vols:
        _logger.warning(
            "seems like you have both volumes and scan on your inputs. Unable to stitch the two at the same time"
        )
    elif has_scans:
        window.setStitchingType(StitchingType.Z_PREPROC)
    elif has_vols:
        window.setStitchingType(StitchingType.Z_POSTPROC)

    window.setAttribute(qt.Qt.WA_DeleteOnClose)
    splash.finish(window)
    exit(app.exec_())


def getinputinfo():
    return "tomwer z-stitching [tomo objects]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.closeAllWindows()  # needed because get a waiting thread behind
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)
